# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import json
from typing import Any, Optional, Tuple

import attrs
import numpy as np
import numpy.typing as npt

from acconeer.exptool.a121._core.utils import attrs_ndarray_eq


@attrs.frozen(kw_only=True)
class Metadata:
    """Metadata

    Represents a superset of the RSS ``processing_metadata``.
    """

    _frame_data_length: int = attrs.field()
    _sweep_data_length: int = attrs.field()
    _subsweep_data_offset: npt.NDArray[np.int_] = attrs.field(eq=attrs_ndarray_eq)
    _subsweep_data_length: npt.NDArray[np.int_] = attrs.field(eq=attrs_ndarray_eq)
    _calibration_temperature: int = attrs.field()
    _tick_period: int = attrs.field()
    _base_step_length_m: float = attrs.field()
    _max_sweep_rate: float = attrs.field()
    _high_speed_mode: Optional[bool] = attrs.field(default=None)

    @property
    def frame_data_length(self) -> int:
        """Number of elements in the frame"""
        return self._frame_data_length

    @property
    def sweep_data_length(self) -> int:
        """Number of elements in the sweep"""
        return self._sweep_data_length

    @property
    def subsweep_data_offset(self) -> npt.NDArray[np.int_]:
        """Offset to the subsweeps data"""
        return self._subsweep_data_offset

    @property
    def subsweep_data_length(self) -> npt.NDArray[np.int_]:
        """Number of elements in the subsweeps"""
        return self._subsweep_data_length

    @property
    def calibration_temperature(self) -> int:
        """Temperature during calibration"""
        return self._calibration_temperature

    @property
    def tick_period(self) -> int:
        """Target tick period if update rate is set, otherwise 0"""
        return self._tick_period

    @property
    def base_step_length_m(self) -> float:
        """Base step length in meter"""
        return self._base_step_length_m

    @property
    def max_sweep_rate(self) -> float:
        """Maximum sweep rate that the sensor can provide for the given configuration"""
        return self._max_sweep_rate

    @property
    def high_speed_mode(self) -> Optional[bool]:
        """Flag indicating if high speed mode is used.
        If true, it means that the sensor has been configured in a way where it
        can optimize its measurements and obtain a high max sweep rate.

        Configuration limitations to enable high speed mode:
        - Continuous sweep mode: off
        - Inter sweep idle state: Ready
        - Subsweeps: 1
        - Profile 3-5

        Note: Available in RSS version > 0.8.0
        """
        return self._high_speed_mode

    @property
    def frame_shape(self) -> Tuple[int, int]:
        """The frame shape this Metadata defines"""

        num_sweeps = self.frame_data_length // self.sweep_data_length
        return (num_sweeps, self.sweep_data_length)

    def to_dict(self) -> dict[str, Any]:
        raw_dict = attrs.asdict(self)
        if self.high_speed_mode is None:
            raw_dict.pop("_high_speed_mode")
        # Remove preceding underscores to be able to recreate the class in from_dict
        return {k[1:] if k.startswith("_") else k: v for k, v in raw_dict.items()}

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Metadata:
        return cls(**d)

    def to_json(self) -> str:
        return json.dumps(self, cls=MetadataEncoder)

    @classmethod
    def from_json(cls, json_str: str) -> Metadata:
        return cls.from_dict(json.loads(json_str, cls=MetadataDecoder))


class MetadataEncoder(json.JSONEncoder):
    """Encoder that transforms a Metadata instance to a serializable
    dict before any json transformation
    """

    def default(self, obj: Any) -> Any:
        if isinstance(obj, Metadata):
            metadata_dict = obj.to_dict()

            # numpy arrays are not serializable.
            # Since they are integer arrays, no precision will be lost
            metadata_dict["subsweep_data_length"] = metadata_dict["subsweep_data_length"].tolist()
            metadata_dict["subsweep_data_offset"] = metadata_dict["subsweep_data_offset"].tolist()
            return metadata_dict

        return super().default(obj)


class MetadataDecoder(json.JSONDecoder):
    """Decoder that post-processes the dict (parsed from json) to better fit Metadata.from_dict"""

    def decode(self, s: str) -> Any:  # type: ignore[override]
        metadata_dict = super().decode(s)

        # post process the parsed dict to have numpy-arrays instead of plain lists
        metadata_dict["subsweep_data_length"] = np.array(metadata_dict["subsweep_data_length"])
        metadata_dict["subsweep_data_offset"] = np.array(metadata_dict["subsweep_data_offset"])

        return metadata_dict



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

from typing import Optional

import pyperclip

from PySide6 import QtGui
from PySide6.QtWidgets import QFrame, QHBoxLayout, QMessageBox, QPushButton, QWidget

from acconeer.exptool.app.new._exceptions import HandledException


class ExceptionWidget(QMessageBox):
    def __init__(
        self,
        parent: QWidget,
        *,
        exc: Exception,
        traceback_str: Optional[str] = None,
        title: str = "Error",
    ) -> None:
        super().__init__(parent)

        self.setIcon(QMessageBox.Icon.Warning)
        self.setStandardButtons(QMessageBox.StandardButton.Ok)

        self.setWindowTitle(title)
        self.setText(str(exc))

        try:
            raise exc
        except HandledException:
            pass
        except Exception:
            self.setInformativeText("<b>Unhandled error - please file a bug</b>")

        if traceback_str:
            self.setDetailedText(traceback_str)
            copy_button = QPushButton(self)
            copy_button.setText("Copy details")
            self.addButton(copy_button, QMessageBox.ButtonRole.ActionRole)
            copy_button.clicked.disconnect()
            copy_button.clicked.connect(self._on_copy_clicked)

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        super().resizeEvent(event)
        self.setFixedWidth(500)

    def _on_copy_clicked(self) -> None:
        detailed_text = self.detailedText()
        if detailed_text:
            pyperclip.copy(detailed_text)

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        self.close()


class VerticalSeparator(QWidget):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.setLayout(QHBoxLayout(self))
        self.layout().setContentsMargins(5, 5, 5, 5)

        frame = QFrame(self)
        frame.setFrameShape(QFrame.Shape.VLine)
        self.layout().addWidget(frame)



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import numpy as np

# Added here to force pyqtgraph to choose PySide
import PySide6  # noqa: F401

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo._utils import estimate_frame_rate
from acconeer.exptool.a121.algo.speed import (
    DetectorConfig,
    Processor,
    ProcessorConfig,
    ProcessorResult,
)
from acconeer.exptool.app.new.ui.stream_tab.plugin_widget import PluginPlotArea


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = a121.Client.open(**a121.get_client_args(args))

    sensor_config = a121.SensorConfig(
        start_point=200,
        step_length=72,
        num_points=1,
        profile=a121.Profile.PROFILE_4,
        sweeps_per_frame=200,
        hwaas=90,
        frame_rate=None,
        sweep_rate=8880.0,
        continuous_sweep_mode=False,
        double_buffering=False,
        inter_frame_idle_state=a121.IdleState.READY,
        inter_sweep_idle_state=a121.IdleState.READY,
        receiver_gain=16,
        enable_tx=True,
        enable_loopback=False,
        phase_enhancement=False,
        prf=a121.PRF.PRF_15_6_MHz,
    )

    SENSOR_ID = 1

    session_config = a121.SessionConfig(
        {SENSOR_ID: sensor_config},
        extended=False,
    )
    estimated_frame_rate = estimate_frame_rate(client, session_config)

    metadata = client.setup_session(sensor_config)

    processor_config = ProcessorConfig(
        threshold=100.0,
        num_segments=4,
    )

    speed_processor = Processor(
        sensor_config=sensor_config,
        metadata=metadata,
        processor_config=processor_config,
    )

    pg_updater = PGUpdater(sensor_config, processor_config, estimated_frame_rate)
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    client.start_session()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        result = client.get_next()
        processed_data = speed_processor.process(result)
        try:
            pg_process.put_data(processed_data)
        except et.PGProccessDiedException:
            break

    print("Disconnecting...")
    pg_process.close()
    client.close()


class PGUpdater:
    def __init__(
        self,
        sensor_config: a121.SensorConfig,
        processor_config: ProcessorConfig,
        estimated_frame_rate: float,
    ):

        self.sensor_config = sensor_config

        self.max_speed = DetectorConfig._get_max_speed(sensor_config.sweep_rate)

        self.n_depths = sensor_config.num_points

        max_update_rate = PluginPlotArea._FPS

        if estimated_frame_rate > max_update_rate:
            plugin_frame_rate = float(max_update_rate)
        else:
            plugin_frame_rate = estimated_frame_rate

        self.history_length_s = 10.0
        self.history_length = int(self.history_length_s * plugin_frame_rate)

        self.time_window_length_s = 3.0
        self.time_window_length_n = int(self.time_window_length_s * plugin_frame_rate)

        self.speed_history = np.zeros(self.history_length)
        self.speed_history_xs = np.array([i for i in range(-self.history_length, 0)])

        n_ticks_to_display = 10
        x_labels = np.linspace(-self.history_length_s, 0, self.history_length)
        all_ticks = [
            (t, "{:.0f}".format(label)) for t, label in zip(self.speed_history_xs, x_labels)
        ]
        subsample_step = self.history_length // n_ticks_to_display
        self.display_ticks = [all_ticks[::subsample_step]]

        self.setup_is_done = False

    def setup(self, win):

        win.setWindowTitle("Acconeer speed detection example")

        self.raw_fft_plot = win.addPlot(row=1, col=0)
        self.raw_fft_plot.setTitle("Frequency data")
        self.raw_fft_plot.setLabel(axis="left", text="Amplitude")
        self.raw_fft_plot.setLabel(axis="bottom", text="Frequency", units="Hz")
        self.raw_fft_plot.addLegend(labelTextSize="10pt")
        self.raw_fft_limits = et.utils.SmoothLimits()
        self.raw_fft_plot.setMenuEnabled(False)
        self.raw_fft_plot.setMouseEnabled(x=False, y=False)
        self.raw_fft_plot.hideButtons()
        self.raw_fft_plot.showGrid(x=True, y=True)
        self.raw_fft_plot.setLogMode(x=False, y=True)
        self.raw_fft_curves = []
        self.raw_fft_smooth_max = et.utils.SmoothMax(self.sensor_config.frame_rate)
        self.raw_thresholds_curves = []

        for i in range(self.n_depths):
            raw_fft_curve = self.raw_fft_plot.plot(pen=et.utils.pg_pen_cycler(i), name="Fft")
            threshold_curve = self.raw_fft_plot.plot(
                pen=et.utils.pg_pen_cycler(i), name="Threshold"
            )
            self.raw_fft_curves.append(raw_fft_curve)
            self.raw_thresholds_curves.append(threshold_curve)

        self.speed_history_plot = win.addPlot(row=0, col=0)
        self.speed_history_plot.setTitle("Speed history")
        self.speed_history_plot.setLabel(axis="left", text="Speed", units="m/s")
        self.speed_history_plot.setLabel(axis="bottom", text="Time", units="Seconds")
        self.speed_history_plot.addLegend(labelTextSize="10pt")
        self.speed_history_curve = self.speed_history_plot.plot(
            pen=None,
            name="speed",
            symbol="o",
            symbolsize=3,
        )

        self.speed_history_plot.setYRange(-self.max_speed, self.max_speed)
        self.speed_history_plot.setXRange(-self.history_length, 0)

        ay = self.speed_history_plot.getAxis("bottom")
        ay.setTicks(self.display_ticks)

        self.speed_html_format = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>"
        )

        self.speed_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0.5),
        )

        self.speed_text_item.setPos(-self.history_length / 2, -self.max_speed / 2)
        brush = et.utils.pg_brush_cycler(1)
        self.speed_history_peak_plot_item = pg.PlotDataItem(
            pen=None, symbol="o", symbolSize=8, symbolBrush=brush, symbolPen="k"
        )
        self.speed_history_plot.addItem(self.speed_history_peak_plot_item)
        self.speed_history_plot.addItem(self.speed_text_item)

        self.speed_text_item.hide()

        self.setup_is_done = True

    def update(self, data: ProcessorResult) -> None:

        psd = data.extra_result.psd
        speed_guess = data.max_speed
        x_speeds = data.extra_result.velocities
        thresholds = data.extra_result.actual_thresholds

        self.speed_history = np.roll(self.speed_history, -1)

        self.speed_history[-1] = speed_guess

        if self.time_window_length_n > 0:
            pos_speed = np.max(self.speed_history[-self.time_window_length_n :])
            pos_ind = int(np.argmax(self.speed_history[-self.time_window_length_n :]))
            neg_speed = np.min(self.speed_history[-self.time_window_length_n :])
            neg_ind = int(np.argmin(self.speed_history[-self.time_window_length_n :]))

            if abs(neg_speed) > abs(pos_speed):
                max_display_speed = neg_speed
                max_display_ind = neg_ind
            else:
                max_display_speed = pos_speed
                max_display_ind = pos_ind
        else:
            max_display_speed = self.speed_history[-1]
            max_display_ind = -1

        if max_display_speed != 0.0:
            speed_text = "Max speed estimate {:.4f} m/s".format(max_display_speed)
            speed_html = self.speed_html_format.format(speed_text)

            self.speed_text_item.setHtml(speed_html)
            self.speed_text_item.show()

            sub_xs = self.speed_history_xs[-self.time_window_length_n :]
            self.speed_history_peak_plot_item.setData(
                [sub_xs[max_display_ind]], [max_display_speed]
            )
        else:
            self.speed_history_peak_plot_item.clear()
            self.speed_text_item.hide()

        display_inds = np.array([i for i, x in enumerate(self.speed_history) if x != 0.0])
        if len(display_inds) > 0:
            display_xs = self.speed_history_xs[display_inds]
            display_data = self.speed_history[display_inds]
        else:
            display_xs = []
            display_data = []
        self.speed_history_curve.setData(display_xs, display_data)

        assert psd is not None
        assert thresholds is not None

        top_max = max(np.max(psd), np.max(thresholds))

        smooth_max_val = np.log10(self.raw_fft_smooth_max.update(top_max))
        self.raw_fft_plot.setYRange(-2, smooth_max_val)
        for i in range(psd.shape[1]):
            self.raw_fft_curves[i].setData(x_speeds, psd[:, i])

            threshold_line = np.full(x_speeds.shape[0], thresholds[i])
            self.raw_thresholds_curves[i].setData(x_speeds, threshold_line)


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import importlib
import logging
from typing import Optional, Tuple

from requests import Response, Session
from requests.cookies import RequestsCookieJar

from PySide6.QtCore import Qt, Signal
from PySide6.QtGui import QCloseEvent, QMovie, QTextCursor
from PySide6.QtWidgets import (
    QCheckBox,
    QDialog,
    QGridLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from acconeer.exptool.app import resources
from acconeer.exptool.app.new.ui.misc import ExceptionWidget
from acconeer.exptool.flash import (
    BIN_FETCH_PROMPT,
    DevLicense,
    save_cookies,
)
from acconeer.exptool.utils import CommDevice

from .threads import AuthThread, FlashThread


_WRONG_CREDENTIALS_MSG = "<font color='red'>Incorrect username (email) or password</font>"

log = logging.getLogger(__name__)


class FlashDialog(QDialog):
    flash_done = Signal()
    opened = Signal()

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        self.setWindowTitle("Flash tool")
        self.setMinimumWidth(250)

        vbox = QVBoxLayout(self)
        vbox.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.loading = QLabel()
        self.loading.setAlignment(Qt.AlignmentFlag.AlignCenter)

        loader_gif = None
        with importlib.resources.path(resources, "loader.gif") as path:
            loader_gif = path

        self.flash_movie = QMovie(str(loader_gif))
        self.loading.setMovie(self.flash_movie)
        vbox.addWidget(self.loading)

        self.flash_label = QLabel(self)
        self.flash_label.setSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.MinimumExpanding
        )
        self.flash_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        vbox.addWidget(self.flash_label)

        self.ok_button = QPushButton(self)
        self.ok_button.setText("OK")
        self.ok_button.setHidden(True)
        self.ok_button.clicked.connect(self._on_ok_click)
        vbox.addWidget(self.ok_button)
        self.setLayout(vbox)

    def flash(self, bin_file: str, flash_device: CommDevice, device_name: Optional[str]) -> None:
        self.flash_thread = FlashThread(bin_file, flash_device, device_name)
        self.flash_thread.started.connect(self._flash_start)
        self.flash_thread.finished.connect(self.flash_thread.deleteLater)
        self.flash_thread.finished.connect(self._flash_stop)
        self.flash_thread.flash_done.connect(self._flash_done)
        self.flash_thread.flash_failed.connect(self._flash_failed)
        self.flash_thread.flash_progress.connect(self._flash_progress)

        self.flash_thread.start()
        self._flashing = True
        self.opened.emit()
        self.exec()

    def _flash_start(self) -> None:
        self.flash_label.setText("Flashing...")
        self.loading.show()
        self.ok_button.setHidden(True)
        self.flash_movie.start()

    def _flash_stop(self) -> None:
        self.flash_movie.stop()
        self.loading.hide()
        self._flashing = False

    def _flash_progress(self, progress: int) -> None:
        if self._flashing:
            self.flash_label.setText(f"Flashing... {progress}%")

    def _flash_done(self) -> None:
        self.flash_done.emit()
        self.flash_label.setText("Flashing done!")
        self.ok_button.setHidden(False)

    def _flash_failed(self, exception: Exception, traceback_str: Optional[str]) -> None:
        self.flash_done.emit()
        self.flash_label.setText("Flashing failed!")
        self.ok_button.setHidden(False)
        ExceptionWidget(self, exc=exception, traceback_str=traceback_str).exec()

    def _on_ok_click(self) -> None:
        self.close()

    def closeEvent(self, event: QCloseEvent) -> None:
        if self._flashing:
            self.flash_thread.terminate()
            self.flash_thread.wait()
        super().closeEvent(event)


class FlashLoginDialog(QDialog):
    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        self.cookies: Optional[RequestsCookieJar] = None
        self.content: Optional[Tuple[bool, Session, Response]] = None
        self.logging_in = False
        self.cookies_accepted = False

        self.setWindowTitle("Acconeer developer login")
        self.setMinimumWidth(450)

        self.fetch_prompt = QTextEdit()
        self.fetch_prompt.setPlainText(BIN_FETCH_PROMPT)
        self.fetch_prompt.setReadOnly(True)

        self.usr_edit = QLineEdit()
        self.usr_edit.setPlaceholderText("<Enter email>")
        self.usr_edit.textChanged.connect(self._on_edit_credentials)

        self.pwd_edit = QLineEdit()
        self.pwd_edit.setPlaceholderText("<Enter password>")
        self.pwd_edit.setEchoMode(QLineEdit.EchoMode.Password)
        self.pwd_edit.textChanged.connect(self._on_edit_credentials)

        self.login_status_label = QLabel()
        self.login_status_label.setHidden(True)

        self.remember_me_box = QCheckBox()
        self.remember_me_box.stateChanged.connect(self._on_remember_me_checked)
        self.remember_me_label = QLabel("Remember me")

        self.login_button = QPushButton("Login")
        self.login_button.setDisabled(True)
        self.login_button.clicked.connect(self._handle_login)

        self.cancel_button = QPushButton("Cancel")
        self.cancel_button.clicked.connect(self.reject)

        self.login_spinner = QLabel()
        self.login_spinner.setHidden(True)

        loader_gif = None
        with importlib.resources.path(resources, "loader.gif") as path:
            loader_gif = path

        self.login_movie = QMovie(str(loader_gif))
        self.login_spinner.setMovie(self.login_movie)

        layout = QGridLayout(self)

        # fmt: off
        # Grid layout:                            row, col, rspan, cspan
        layout.addWidget(self.fetch_prompt,       0,   0,   1,     12)    # noqa: E241
        layout.addWidget(self.usr_edit,           1,   0,   1,     12)    # noqa: E241
        layout.addWidget(self.pwd_edit,           2,   0,   1,     12)    # noqa: E241
        layout.addWidget(self.remember_me_box,    3,   0,   1,     1)     # noqa: E241
        layout.addWidget(self.remember_me_label,  3,   1,   1,     11)    # noqa: E241
        layout.addWidget(self.login_status_label, 4,   0,   1,     12)    # noqa: E241
        layout.addWidget(self.login_button,       5,   0,   1,     6)     # noqa: E241
        layout.addWidget(self.login_spinner,      5,   6,   1,     3)     # noqa: E241
        layout.addWidget(self.cancel_button,      5,   9,   1,     3)     # noqa: E241
        # fmt: on

        self.setLayout(layout)

        self.usr_edit.setFocus()

    def _on_edit_credentials(self) -> None:
        if self.usr_edit.text() == "" or self.pwd_edit.text() == "":
            self.login_button.setEnabled(False)
        else:
            self.login_button.setEnabled(True)

    def _on_remember_me_checked(self) -> None:
        if self.remember_me_box.isChecked() and not self.cookies_accepted:
            cookie_dialog = CookieConsentDialog(self)
            cookie_dialog.exec()
            self.cookies_accepted = cookie_dialog.result() == QDialog.DialogCode.Accepted
            if not self.cookies_accepted:
                self.remember_me_box.setCheckState(Qt.CheckState.Unchecked)

    def _handle_login(self) -> None:
        self.login_status_label.setHidden(True)

        usr = self.usr_edit.text()
        pwd = self.pwd_edit.text()

        self.login_thread = AuthThread(usr, pwd, do_login=True)
        self.login_thread.started.connect(self._login_start)
        self.login_thread.finished.connect(self.login_thread.deleteLater)
        self.login_thread.finished.connect(self._login_stop)
        self.login_thread.auth_done.connect(self._login_done)
        self.login_thread.auth_failed.connect(self._login_failed)

        self.login_thread.start()
        self.logging_in = True

    def _login_start(self) -> None:
        self.setEnabled(False)
        self._show_login_progress()

    def _login_stop(self) -> None:
        self.logging_in = False
        self._hide_login_progress()
        self.setEnabled(True)

    def _login_done(
        self, auth_info: Tuple[RequestsCookieJar, Tuple[bool, Session, Response]]
    ) -> None:

        self.cookies, self.content = auth_info

        self.login_status_label.setHidden(True)

        if self.remember_me_box.isChecked():
            save_cookies(self.cookies)

        self.accept()

    def _login_failed(self) -> None:
        self.login_status_label.setText(_WRONG_CREDENTIALS_MSG)
        self.login_status_label.setHidden(False)

    def _show_login_progress(self) -> None:
        self.login_movie.start()
        self.login_spinner.setHidden(False)

    def _hide_login_progress(self) -> None:
        self.login_movie.stop()
        self.login_spinner.setHidden(True)

    def get_auth_info(
        self,
    ) -> (Tuple[Optional[RequestsCookieJar], Optional[Tuple[bool, Session, Response]]]):
        return self.cookies, self.content

    def closeEvent(self, event: QCloseEvent) -> None:
        if self.logging_in:
            self.login_thread.terminate()
            self.login_thread.wait()
        super().closeEvent(event)


class CookieConsentDialog(QDialog):
    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        self.setWindowTitle("Cookie consent")

        layout = QGridLayout(self)

        self.consent_label = QLabel(
            "We use cookies to optimize the service of our applications", self
        )
        self.empty_row = QLabel("", self)

        self.cookie_policy_label = QLabel(
            "<a href='https://acconeer.com/cookie-policy-eu/'>Cookie Policy</a>", self
        )
        self.cookie_policy_label.setOpenExternalLinks(True)

        self.privacy_statement_label = QLabel(
            "<a href='https://developer.acconeer.com/privacy-policy/'>Privacy Statement</a>",
            self,
        )
        self.privacy_statement_label.setOpenExternalLinks(True)

        self.accept_button = QPushButton("Accept", self)
        self.accept_button.clicked.connect(self.accept)

        self.deny_button = QPushButton("Deny", self)
        self.deny_button.clicked.connect(self.reject)

        # fmt: off
        # Grid layout:                                 row, col, rspan, cspan
        layout.addWidget(self.consent_label,           0,   0,   1,     4)    # noqa: E241
        layout.addWidget(self.empty_row,               1,   0,   1,     4)    # noqa: E241
        layout.addWidget(self.cookie_policy_label,     2,   0,   1,     4)    # noqa: E241
        layout.addWidget(self.privacy_statement_label, 3,   0,   1,     4)    # noqa: E241
        layout.addWidget(self.accept_button,           4,   0,   1,     2)    # noqa: E241
        layout.addWidget(self.deny_button,             4,   2,   1,     2)    # noqa: E241
        # fmt: on

        self.setLayout(layout)


class LicenseAgreementDialog(QDialog):
    def __init__(self, license: DevLicense, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        self.setMinimumWidth(800)
        self.setMinimumHeight(600)

        layout = QGridLayout(self)

        self.license_text = QTextEdit(self)
        self.license_text.setReadOnly(True)

        self.accept_box = QCheckBox(self)
        self.accept_box.stateChanged.connect(self._on_accept_check)
        self.accept_label = QLabel("I understand the terms and conditions of the license", self)

        self.accept_button = QPushButton("Accept", self)
        self.accept_button.setEnabled(False)
        self.accept_button.clicked.connect(self.accept)

        self.cancel_button = QPushButton("Cancel", self)
        self.cancel_button.clicked.connect(self.reject)

        # fmt: off
        # Grid layout:                       row, col, rspan, cspan
        layout.addWidget(self.license_text,  0,   0,   10,    12)    # noqa: E241
        layout.addWidget(self.accept_box,    10,  0,   1,     1)     # noqa: E241
        layout.addWidget(self.accept_label,  10,  1,   1,     11)    # noqa: E241
        layout.addWidget(self.accept_button, 11,  0,   1,     6)     # noqa: E241
        layout.addWidget(self.cancel_button, 11,  6,   1,     6)     # noqa: E241
        # fmt: on

        self.setLayout(layout)

        self.set_license_text(license)

    def _on_accept_check(self) -> None:
        self.accept_button.setEnabled(self.accept_box.isChecked())

    def set_license_text(self, license: DevLicense) -> None:
        self.setWindowTitle(license.get_header())

        self.license_text.insertHtml(license.get_subheader_element())

        content = license.get_content_elements()
        for paragraph in content:
            self.license_text.insertHtml(paragraph)
            self.license_text.insertHtml("<br><br>")

        self.license_text.moveCursor(QTextCursor.MoveOperation.Start)


class UserMessageDialog(QDialog):
    def __init__(
        self,
        title: str,
        message: Optional[str],
        confirmation: str,
        parent: Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent)

        self.setMinimumWidth(450)
        self.setMinimumHeight(200)

        layout = QGridLayout(self)

        self.message_text = QTextEdit(self)
        self.message_text.setReadOnly(True)

        self.ok_button = QPushButton(confirmation, self)
        self.ok_button.setEnabled(True)
        self.ok_button.clicked.connect(self._on_ok)

        # fmt: off
        # Grid layout:                      row, col, rspan, cspan
        layout.addWidget(self.message_text, 0,   0,   4,     12)    # noqa: E241
        layout.addWidget(self.ok_button,    4,   0,   1,     2)     # noqa: E241
        # fmt: on

        self.setLayout(layout)

        self.setWindowTitle(title)
        if message is not None:
            self.set_message(message)

    def set_message(self, message: str) -> None:
        self.message_text.clear()
        self.message_text.insertHtml(message)
        self.message_text.moveCursor(QTextCursor.MoveOperation.Start)

    def _on_ok(self) -> None:
        self.accept()



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

from typing import Optional, Union

from acconeer.exptool.a121._core.entities import (
    ClientInfo,
    Metadata,
    Result,
    SensorCalibration,
    SessionConfig,
)
from acconeer.exptool.a121._core.mediators import Recorder
from acconeer.exptool.a121._core.utils import unextend

from .client import Client, ClientError


class CommonClient(Client):
    _client_info: ClientInfo
    _calibrations_provided: dict[int, bool]
    _metadata: Optional[list[dict[int, Metadata]]]
    _recorder: Optional[Recorder]
    _sensor_calibrations: Optional[dict[int, SensorCalibration]]
    _session_config: Optional[SessionConfig]
    _session_is_started: bool

    def _assert_connected(self) -> None:
        if not self.connected:
            raise ClientError("Client is not connected.")

    def _assert_session_setup(self) -> None:
        self._assert_connected()
        if not self.session_is_setup:
            raise ClientError("Session is not set up.")

    def _assert_session_started(self) -> None:
        self._assert_session_setup()
        if not self.session_is_started:
            raise ClientError("Session is not started.")

    def __init__(self, client_info: ClientInfo) -> None:
        self._client_info = client_info
        self._calibrations_provided = {}
        self._metadata = None
        self._recorder = None
        self._sensor_calibrations = None
        self._session_is_started = False
        self._session_config = None

    def attach_recorder(self, recorder: Recorder) -> None:
        if self.session_is_started:
            raise ClientError("Cannot attach a recorder when session is started.")

        if not self.connected:
            raise ClientError("Cannot attach a recorder to a closed client")

        if self._recorder is not None:
            raise ClientError(
                "Client already has a recorder attached. "
                + "Try detaching the current recorder before attaching a new recorder."
            )

        self._recorder = recorder
        self._recorder._start(
            client_info=self.client_info,
            server_info=self.server_info,
        )

    def detach_recorder(self) -> Optional[Recorder]:
        if self.session_is_started:
            raise ClientError("Cannot detach a recorder when session is started.")

        if not self.connected:
            raise ClientError("Cannot detach a recorder from a closed client")

        if self._recorder is None:
            return None
        else:
            previously_attached_recorder = self._recorder
            self._recorder = None
            return previously_attached_recorder

    def _recorder_start_session(self) -> None:
        if self._recorder is not None:
            calibrations_provided: Optional[dict[int, bool]] = self.calibrations_provided
            try:
                calibrations = self.calibrations
            except ClientError:
                calibrations = None
                calibrations_provided = None

            self._recorder._start_session(
                session_config=self.session_config,
                extended_metadata=self.extended_metadata,
                calibrations=calibrations,
                calibrations_provided=calibrations_provided,
            )

    def _recorder_stop_session(self) -> None:
        if self._recorder is not None:
            self._recorder._stop_session()

    def _recorder_sample(self, extended_results: list[dict[int, Result]]) -> None:
        if self._recorder is not None:
            self._recorder._sample(extended_results)

    def _return_results(
        self, extended_results: list[dict[int, Result]]
    ) -> Union[Result, list[dict[int, Result]]]:
        if self.session_config.extended:
            return extended_results
        else:
            return unextend(extended_results)

    @property
    def session_is_setup(self) -> bool:
        return self._metadata is not None

    @property
    def session_is_started(self) -> bool:
        return self._session_is_started

    @property
    def client_info(self) -> ClientInfo:
        return self._client_info

    @property
    def session_config(self) -> SessionConfig:
        self._assert_session_setup()
        assert self._session_config is not None  # Should never happen if session is setup
        return self._session_config

    @property
    def extended_metadata(self) -> list[dict[int, Metadata]]:
        self._assert_session_setup()
        assert self._metadata is not None  # Should never happen if session is setup
        return self._metadata

    @property
    def calibrations(self) -> dict[int, SensorCalibration]:
        self._assert_session_setup()

        if not self._sensor_calibrations:
            raise ClientError("Server did not provide calibration")

        return self._sensor_calibrations

    @property
    def calibrations_provided(self) -> dict[int, bool]:
        return self._calibrations_provided



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import json
import logging
import typing as t
from pathlib import Path

import attrs
import typing_extensions as te

from PySide6.QtCore import QSize, Qt, Signal
from PySide6.QtGui import QAction, QIcon, QMouseEvent
from PySide6.QtWidgets import (
    QDockWidget,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QMessageBox,
    QToolBar,
    QToolButton,
    QWidget,
)

import acconeer.exptool
from acconeer.exptool import a121
from acconeer.exptool.a121.algo import distance, presence
from acconeer.exptool.app.new.storage import get_config_dir
from acconeer.exptool.app.new.ui.icons import CHART_BAR, CHART_LINE, CLOSE, COG, INFO, MEMORY
from acconeer.exptool.utils import get_module_version

from .animation import run_blink_animation
from .event_system import EventBroker
from .services import (
    distance_config_input,
    memory_breakdown_output,
    power_consumption_vs_rate_output,
    power_curve_output,
    presence_config_input,
    session_config_input,
)


_QWidgetT = t.TypeVar("_QWidgetT", bound=t.Type[QWidget])


log = logging.getLogger(__name__)

_ALLOWED_DOCK_AREA = Qt.DockWidgetArea.BottomDockWidgetArea

_DATA_IS_APPROXIMATE_DESCRIPTION = "\n\n".join(
    [
        "All data and numbers presented in this tab are approximations. "
        + "The goal is to let you experiment with configurations and do comparisons.",
        "Use this tool as a way to quickly narrow down which configuration could satisfy your "
        + "constraints before you do measurements to get real numbers.",
    ]
)


_TAB_BRIEF_DESCRIPTION = "\n\n".join(
    [
        "This tab allows you to experiment and visualize resource "
        + "requirements for different configurations. ",
        'The top bar is a palette with all the "blocks" you can use. '
        + "Click on them to add an additional block of that type.",
        "Some blocks have an id in [brackets] in the title bar. "
        + "This is used to connect inputs and outputs.",
        "You can hover a block's info icon for a longer description "
        + "of what the block does and how you may use it.",
    ]
)


@attrs.mutable
class _UserUnderstandings:
    """
    Simple class that keeps track of what the user understands.

    Used to control pop-ups with information when opening the resource tab.
    """

    _PATH: t.ClassVar[Path] = get_config_dir() / "user_understandings.json"

    data_is_approximate: bool
    tab_brief: bool
    lib_version: str

    @classmethod
    def require(cls) -> te.Self:
        """Returns a persisted instance or creates a new instance"""
        try:
            with cls._PATH.open("r") as file:
                return cls(**json.load(file))
        except (FileNotFoundError, json.JSONDecodeError):
            return cls(
                data_is_approximate=False,
                tab_brief=False,
                lib_version=get_module_version(acconeer.exptool),
            )

    def save(self) -> None:
        with self._PATH.open("w") as file:
            return json.dump(attrs.asdict(self), file)


class _TitleBarLabel(QLabel):
    def mouseDoubleClickEvent(self, event: QMouseEvent) -> None:
        event.ignore()

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        event.ignore()

    def mousePressEvent(self, event: QMouseEvent) -> None:
        event.ignore()

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        event.ignore()


class _TitleBarWidget(QWidget):
    sig_exit_clicked = Signal()

    def __init__(
        self,
        title: str,
        tooltip: str,
        parent: QWidget,
    ) -> None:
        super().__init__(parent=parent)
        self.setAttribute(Qt.WidgetAttribute.WA_StyledBackground)
        self.setObjectName("_TitleBarWidget")
        self.setStyleSheet("#_TitleBarWidget { background: #ebebeb; }")

        title_label = _TitleBarLabel(title)
        title_label.setTextFormat(Qt.TextFormat.RichText)

        info_icon = QToolButton()
        info_icon.setIcon(INFO())
        info_icon.setIconSize(QSize(16, 16))
        info_icon.setDisabled(True)
        info_icon.setToolTip(tooltip)

        close_button = QToolButton()
        close_button.setIcon(CLOSE())
        close_button.clicked.connect(self.sig_exit_clicked.emit)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(title_label, stretch=0)
        layout.addWidget(info_icon, stretch=0, alignment=Qt.AlignmentFlag.AlignLeft)
        layout.addWidget(close_button, stretch=1, alignment=Qt.AlignmentFlag.AlignRight)

        self.setLayout(layout)

    def mouseDoubleClickEvent(self, event: QMouseEvent) -> None:
        event.ignore()

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        event.ignore()

    def mousePressEvent(self, event: QMouseEvent) -> None:
        event.ignore()

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        event.ignore()


class _DockWidget(QDockWidget):
    def __init__(
        self,
        title: str,
        tooltip: str,
        widget: QWidget,
        close_hook: t.Callable[[], None],
        parent: QWidget,
    ) -> None:
        super().__init__(parent=parent)
        self.setWindowTitle(title)

        self.setWidget(widget)
        self.setAllowedAreas(_ALLOWED_DOCK_AREA)
        self._close_hook = close_hook

        title_bar = _TitleBarWidget(title, tooltip, self)
        title_bar.sig_exit_clicked.connect(close_hook)
        title_bar.sig_exit_clicked.connect(self.close)
        self.setTitleBarWidget(title_bar)


class ResourceMainWidget(QMainWindow):
    def __init__(self) -> None:
        super().__init__()

        self._broker = EventBroker()

        # default is AnimatedDocks | AllowTabbedDocks
        self.setDockOptions(
            QMainWindow.DockOption.AllowNestedDocks | QMainWindow.DockOption.AnimatedDocks
        )

        toolbar = QToolBar()
        toolbar.setFixedHeight(64)
        toolbar.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        toolbar.setIconSize(QSize(24, 24))
        toolbar.setFloatable(False)
        toolbar.setMovable(False)

        self._add_spawn_buttons(toolbar)
        self.setCentralWidget(toolbar)

        # Default loadout
        power_curve = self._create_power_curve_output()
        self._add_dock_widget(power_curve)
        self._add_dock_widget_below_target(
            target=power_curve,
            new_widget=self._create_memory_breakdown_output(),
        )
        self._add_dock_widget_right_of_target(
            target=power_curve,
            new_widget=self._create_power_consumption_vs_rate_output(),
        )
        self.spawn_input_block(a121.SessionConfig(), animate=False)

    def showEvent(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().showEvent(*args, **kwargs)

        user_understandings = _UserUnderstandings.require()

        if not user_understandings.data_is_approximate:
            pressed_button = QMessageBox.information(
                self, "Data is approximate", _DATA_IS_APPROXIMATE_DESCRIPTION
            )
            user_understandings.data_is_approximate = (
                pressed_button == QMessageBox.StandardButton.Ok
            )

        if not user_understandings.tab_brief:
            pressed_button = QMessageBox.information(
                self, "Resource Calculator", _TAB_BRIEF_DESCRIPTION
            )
            user_understandings.tab_brief = pressed_button == QMessageBox.StandardButton.Ok

        user_understandings.save()

    def spawn_input_block(self, config: t.Any, animate: bool = True) -> None:
        if isinstance(config, a121.SessionConfig):
            dock_widget = self._create_session_config_input(config)
        elif isinstance(config, distance.DetectorConfig):
            dock_widget = self._create_distance_config_input(config)
        elif isinstance(config, presence.DetectorConfig):
            dock_widget = self._create_presence_config_input(config)
        else:
            return

        self._add_dock_widget(dock_widget)
        if animate:
            run_blink_animation(dock_widget)

    def _add_dock_widget(self, dockwidget: QDockWidget) -> None:
        super().addDockWidget(_ALLOWED_DOCK_AREA, dockwidget)

    def _add_dock_widget_below_target(self, target: QDockWidget, new_widget: QDockWidget) -> None:
        super().splitDockWidget(target, new_widget, Qt.Orientation.Vertical)

    def _add_dock_widget_right_of_target(
        self, target: QDockWidget, new_widget: QDockWidget
    ) -> None:
        super().splitDockWidget(target, new_widget, Qt.Orientation.Horizontal)

    def _create_action(
        self,
        icon: QIcon,
        button_label: str,
        on_trigger: t.Callable[[], t.Any],
    ) -> QAction:
        action = QAction(icon, button_label, parent=self)
        action.triggered.connect(on_trigger)
        return action

    def _add_spawn_buttons(self, toolbar: QToolBar) -> None:
        actions = [
            self._create_action(
                COG(),
                button_label="Sparse IQ config",
                on_trigger=lambda: self.spawn_input_block(a121.SessionConfig(), animate=False),
            ),
            self._create_action(
                COG(),
                "Distance config",
                on_trigger=lambda: self.spawn_input_block(
                    distance.DetectorConfig(), animate=False
                ),
            ),
            self._create_action(
                COG(),
                "Presence config",
                on_trigger=lambda: self.spawn_input_block(
                    presence.DetectorConfig(), animate=False
                ),
            ),
            self._create_action(
                CHART_BAR(),
                button_label="Power curve",
                on_trigger=lambda: self._add_dock_widget(self._create_power_curve_output()),
            ),
            self._create_action(
                CHART_LINE(),
                button_label="Power consumption vs. Rate",
                on_trigger=lambda: self._add_dock_widget(
                    self._create_power_consumption_vs_rate_output()
                ),
            ),
            self._create_action(
                MEMORY(),
                button_label="Memory breakdown",
                on_trigger=lambda: self._add_dock_widget(self._create_memory_breakdown_output()),
            ),
        ]

        for action in actions:
            toolbar.addAction(action)

    def _create_session_config_input(self, config: a121.SessionConfig) -> _DockWidget:
        service = session_config_input.SessionConfigInput(self._broker, config)
        return _DockWidget(
            service.window_title,
            service.description,
            service,
            service.uninstall_function,
            parent=self,
        )

    def _create_power_curve_output(self) -> _DockWidget:
        service = power_curve_output.EnergyRegionOutput(self._broker)
        return _DockWidget(
            service.window_title,
            service.description,
            service,
            service.uninstall_function,
            parent=self,
        )

    def _create_power_consumption_vs_rate_output(self) -> _DockWidget:
        service = power_consumption_vs_rate_output.PowerConsumptionVsRateOutput(self._broker)
        return _DockWidget(
            service.window_title,
            service.description,
            service,
            service.uninstall_function,
            parent=self,
        )

    def _create_memory_breakdown_output(self) -> _DockWidget:
        service = memory_breakdown_output.MemoryBreakdownOutput(self._broker)
        return _DockWidget(
            service.window_title,
            service.description,
            service,
            service.uninstall_function,
            parent=self,
        )

    def _create_distance_config_input(self, config: distance.DetectorConfig) -> _DockWidget:
        service = distance_config_input.DistanceConfigInput(self._broker, config)
        return _DockWidget(
            service.window_title,
            service.description,
            service,
            service.uninstall_function,
            parent=self,
        )

    def _create_presence_config_input(self, config: presence.DetectorConfig) -> _DockWidget:
        service = presence_config_input.PresenceConfigInput(self._broker, config)
        return _DockWidget(
            service.window_title,
            service.description,
            service,
            service.uninstall_function,
            parent=self,
        )



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import abc
import logging
from pathlib import Path
from typing import Callable, Generic, Optional, TypeVar

import h5py

from acconeer.exptool import a121
from acconeer.exptool.app.new import (
    ApplicationClient,
    BackendPlugin,
    GeneralMessage,
    HandledException,
    Message,
    PluginGeneration,
    PluginState,
    PluginStateMessage,
    ViewPluginBase,
    get_temp_h5_path,
    is_task,
)


log = logging.getLogger(__name__)


T = TypeVar("T")


class A121BackendPluginBase(Generic[T], BackendPlugin[T]):
    _live_client: Optional[a121.Client]
    _replaying_client: Optional[a121.Client]
    _opened_record: Optional[a121.H5Record]
    _started: bool = False
    _recorder: Optional[a121.H5Recorder] = None

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ) -> None:
        super().__init__(callback, generation, key)
        self._live_client = None
        self._replaying_client = None
        self._opened_record = None

    @is_task
    def load_from_file(self, *, path: Path) -> None:
        try:
            self._opened_record = a121.H5Record(h5py.File(path, mode="r"))

            # This is to be able to replay files recorded before v7.0.0
            # when there were only one session per file
            if self._opened_record.num_sessions <= 1:
                replaying_client = a121._ReplayingClient(self._opened_record, cycled_session_idx=0)
            else:
                replaying_client = a121._ReplayingClient(self._opened_record)

            self.load_from_record_setup(record=self._opened_record)
        except Exception as exc:
            self._opened_record = None
            self._replaying_client = None

            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))
            raise HandledException("Could not load from file") from exc

        self._replaying_client = ApplicationClient.wrap(replaying_client, self.callback)

        self.start_session(with_recorder=False)

        self.send_status_message(f"<b>Replaying from {path.name}</b>")
        self.broadcast()

    @abc.abstractmethod
    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        pass

    @is_task
    def load_from_cache(self) -> None:
        try:
            with self.h5_cache_file() as f:
                self._load_from_cache(f)
        except FileNotFoundError:
            pass

        self._sync_sensor_ids()
        self.broadcast()

    @abc.abstractmethod
    def _load_from_cache(self, file: h5py.File) -> None:
        pass

    @abc.abstractmethod
    def get_next(self) -> None:
        pass

    def idle(self) -> bool:
        if self._started:
            if self.client is None:
                raise RuntimeError("Client is not attached. Can not 'get_next'.")

            try:
                self.get_next()
            except a121._StopReplay:
                self.stop_session()
                return True
            except Exception as exc:
                try:
                    self.stop_session()
                except Exception:
                    pass

                raise HandledException("Failed to get_next") from exc

            return True
        else:
            return False

    @property
    def client(self) -> Optional[a121.Client]:
        if self._replaying_client is not None:
            return self._replaying_client

        return self._live_client

    @abc.abstractmethod
    def _sync_sensor_ids(self) -> None:
        pass

    def attach_client(self, *, client: a121.Client) -> None:
        self._live_client = ApplicationClient.wrap(client, self.callback)
        self._sync_sensor_ids()
        self.broadcast()

    def detach_client(self) -> None:
        self._live_client = None

    @abc.abstractmethod
    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        pass

    @abc.abstractmethod
    def end_session(self) -> None:
        pass

    @is_task
    def start_session(self, *, with_recorder: bool = True) -> None:
        if self._started:
            raise RuntimeError

        if self.client is None:
            raise RuntimeError("Client is not attached. Can not 'start'.")

        if not self.client.connected:
            raise RuntimeError("Client is not connected. Can not 'start'.")

        self.callback(GeneralMessage(name="saveable_file", data=None))
        if with_recorder:
            self._recorder = a121.H5Recorder(get_temp_h5_path())
        else:
            self._recorder = None

        self.callback(PluginStateMessage(state=PluginState.LOADED_STARTING))
        try:
            self._start_session(self._recorder)
        except Exception as exc:
            recorder = self.client.detach_recorder()
            if recorder is not None:
                recorder.close()
            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))
            raise HandledException("Could not start") from exc

        self._started = True

        self.broadcast()

        self.callback(PluginStateMessage(state=PluginState.LOADED_BUSY))

    @is_task
    def stop_session(self) -> None:
        if not self._started:
            raise RuntimeError

        self.callback(PluginStateMessage(state=PluginState.LOADED_STOPPING))
        try:
            self.end_session()
        except Exception as exc:
            raise HandledException("Failure when stopping session") from exc
        finally:
            if self._recorder is not None:
                assert self._recorder.path is not None
                path = Path(self._recorder.path)
                self.callback(GeneralMessage(name="saveable_file", data=path))
                self._recorder = None

            self._started = False
            self.broadcast()
            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))

        if self._opened_record:
            self._opened_record.close()
            self._opened_record = None
            self._replaying_client = None
            self._sync_sensor_ids()

    def teardown(self) -> None:
        try:
            with self.h5_cache_file(write=True) as file:
                self.save_to_cache(file)
        except Exception:
            log.warning("Could not write to cache")

        self.detach_client()

    @abc.abstractmethod
    def save_to_cache(self, file: h5py.File) -> None:
        pass


class A121ViewPluginBase(ViewPluginBase):
    def _send_start_request(self) -> None:
        A121BackendPluginBase.start_session.rpc(
            self.app_model.put_task,
            with_recorder=self.app_model.recording_enabled,
        )

    def _send_stop_request(self) -> None:
        A121BackendPluginBase.stop_session.rpc(self.app_model.put_task)



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

from typing import List, Tuple

import numpy as np
import numpy.typing as npt

from PySide6 import QtCore

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo.presence import Detector
from acconeer.exptool.a121.algo.smart_presence._ref_app import (
    PresenceWakeUpConfig,
    PresenceZoneConfig,
    RefApp,
    RefAppConfig,
    RefAppResult,
    _Mode,
)


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = a121.Client.open(**a121.get_client_args(args))

    ref_app_config = RefAppConfig(
        wake_up_mode=True,
        wake_up_config=PresenceWakeUpConfig(
            start_m=1.0,
            end_m=3.0,
            num_zones=5,
            num_zones_for_wake_up=2,
        ),
        nominal_config=PresenceZoneConfig(
            start_m=1.0,
            end_m=3.0,
            num_zones=3,
        ),
    )

    ref_app = RefApp(client=client, sensor_id=1, ref_app_config=ref_app_config)
    ref_app.start()

    nominal_sensor_config = Detector._get_sensor_config(ref_app.nominal_detector_config)
    distances = np.linspace(
        ref_app_config.nominal_config.start_m,
        ref_app_config.nominal_config.end_m,
        nominal_sensor_config.num_points,
    )
    nominal_zone_limits = ref_app.ref_app_processor.create_zones(
        distances, ref_app_config.nominal_config.num_zones
    )

    pg_updater = PGUpdater(
        ref_app_config,
        ref_app.ref_app_context.wake_up_detector_context.estimated_frame_rate,
        nominal_zone_limits,
        ref_app.ref_app_processor.zone_limits,
    )
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        ref_app_result = ref_app.get_next()
        if ref_app_result.presence_detected:
            print(f"Presence in zone {ref_app_result.max_presence_zone}")
        else:
            print("No presence")
        try:
            pg_process.put_data(ref_app_result)
        except et.PGProccessDiedException:
            break

    ref_app.stop()

    print("Disconnecting...")
    client.close()


class PGUpdater:
    def __init__(
        self,
        ref_app_config: RefAppConfig,
        estimated_frame_rate: float,
        nominal_zone_limits: npt.NDArray[np.float_],
        wake_up_zone_limits: npt.NDArray[np.float_],
    ) -> None:
        self.ref_app_config = ref_app_config
        self.nominal_config = ref_app_config.nominal_config
        self.wake_up_config = ref_app_config.wake_up_config

        self.show_all_detected_zones = ref_app_config.show_all_detected_zones
        self.nominal_zone_limits = nominal_zone_limits
        self.wake_up_zone_limits = wake_up_zone_limits
        self.estimated_frame_rate = estimated_frame_rate

        self.history_length_s = 5
        self.time_fifo: List[float] = []
        self.intra_fifo: List[float] = []
        self.inter_fifo: List[float] = []

        self.intra_limit_lines = []
        self.inter_limit_lines = []

        self.setup_is_done = False

    def setup(self, win):
        win.setWindowTitle("Acconeer smart presence example")

        # Intra presence history plot

        self.intra_hist_plot = win.addPlot(
            row=0,
            col=0,
            title="Intra presence history (fast motions)",
        )
        self.intra_hist_plot.setMenuEnabled(False)
        self.intra_hist_plot.setMouseEnabled(x=False, y=False)
        self.intra_hist_plot.hideButtons()
        self.intra_hist_plot.showGrid(x=True, y=True)
        self.intra_hist_plot.setLabel("bottom", "Time (s)")
        self.intra_hist_plot.setLabel("left", "Score")
        self.intra_hist_plot.setXRange(-self.history_length_s, 0)
        self.intra_history_smooth_max = et.utils.SmoothMax(self.estimated_frame_rate)
        self.intra_hist_plot.setYRange(0, 10)
        if not self.nominal_config.intra_enable:
            intra_color = et.utils.color_cycler(1)
            intra_color = f"{intra_color}50"
            self.nominal_intra_dashed_pen = pg.mkPen(
                intra_color, width=2.5, style=QtCore.Qt.DashLine
            )
            self.nominal_intra_pen = pg.mkPen(intra_color, width=2)
        else:
            self.nominal_intra_dashed_pen = et.utils.pg_pen_cycler(1, width=2.5, style="--")
            self.nominal_intra_pen = et.utils.pg_pen_cycler(1)

        self.intra_hist_curve = self.intra_hist_plot.plot(pen=self.nominal_intra_pen)
        limit_line = pg.InfiniteLine(angle=0, pen=self.nominal_intra_dashed_pen)
        self.intra_hist_plot.addItem(limit_line)
        self.intra_limit_lines.append(limit_line)

        for line in self.intra_limit_lines:
            line.setPos(self.nominal_config.intra_detection_threshold)

        # Inter presence history plot

        self.inter_hist_plot = win.addPlot(
            row=0,
            col=1,
            title="Inter presence history (slow motions)",
        )
        self.inter_hist_plot.setMenuEnabled(False)
        self.inter_hist_plot.setMouseEnabled(x=False, y=False)
        self.inter_hist_plot.hideButtons()
        self.inter_hist_plot.showGrid(x=True, y=True)
        self.inter_hist_plot.setLabel("bottom", "Time (s)")
        self.inter_hist_plot.setLabel("left", "Score")
        self.inter_hist_plot.setXRange(-self.history_length_s, 0)
        self.inter_history_smooth_max = et.utils.SmoothMax(self.estimated_frame_rate)
        self.inter_hist_plot.setYRange(0, 10)
        if not self.nominal_config.inter_enable:
            inter_color = et.utils.color_cycler(0)
            inter_color = f"{inter_color}50"
            self.nominal_inter_dashed_pen = pg.mkPen(
                inter_color, width=2.5, style=QtCore.Qt.DashLine
            )
            self.nominal_inter_pen = pg.mkPen(inter_color, width=2)
        else:
            self.nominal_inter_pen = et.utils.pg_pen_cycler(0)
            self.nominal_inter_dashed_pen = et.utils.pg_pen_cycler(0, width=2.5, style="--")

        self.inter_hist_curve = self.inter_hist_plot.plot(pen=self.nominal_inter_pen)
        limit_line = pg.InfiniteLine(angle=0, pen=self.nominal_inter_dashed_pen)
        self.inter_hist_plot.addItem(limit_line)
        self.inter_limit_lines.append(limit_line)

        for line in self.inter_limit_lines:
            line.setPos(self.nominal_config.inter_detection_threshold)

        # Sector plot

        if self.ref_app_config.wake_up_mode:
            title = (
                "Nominal config<br>"
                "Detection type: fast (orange), slow (blue), both (green)<br>"
                "Green background indicates active"
            )
        else:
            title = "Nominal config<br>" "Detection type: fast (orange), slow (blue), both (green)"

        self.nominal_sector_plot, self.nominal_sectors = self.create_sector_plot(
            title,
            self.ref_app_config.nominal_config.num_zones,
            self.nominal_config.start_m,
            self.nominal_zone_limits,
        )

        if not self.ref_app_config.wake_up_mode:
            sublayout = win.addLayout(row=1, col=0, colspan=2)
            sublayout.layout.setColumnStretchFactor(0, 2)
            sublayout.addItem(self.nominal_sector_plot, row=0, col=0)
        else:
            assert self.wake_up_config is not None
            sublayout = win.addLayout(row=1, col=0, colspan=2)
            sublayout.addItem(self.nominal_sector_plot, row=0, col=1)

            title = (
                "Wake up config<br>"
                "Detection type: fast (orange), slow (blue), both (green),<br>"
                "lingering (light grey)<br>"
                "Green background indicates active"
            )
            self.wake_up_sector_plot, self.wake_up_sectors = self.create_sector_plot(
                title,
                self.wake_up_config.num_zones,
                self.wake_up_config.start_m,
                self.wake_up_zone_limits,
            )

            sublayout.addItem(self.wake_up_sector_plot, row=0, col=0)

            if self.wake_up_config.intra_enable:
                self.wake_up_intra_dashed_pen = et.utils.pg_pen_cycler(1, width=2.5, style="--")
                self.wake_up_intra_pen = et.utils.pg_pen_cycler(1)
            else:
                intra_color = et.utils.color_cycler(1)
                intra_color = f"{intra_color}50"
                self.wake_up_intra_dashed_pen = pg.mkPen(
                    intra_color, width=2.5, style=QtCore.Qt.DashLine
                )
                self.wake_up_intra_pen = pg.mkPen(intra_color, width=2)

            if self.wake_up_config.inter_enable:
                self.wake_up_inter_pen = et.utils.pg_pen_cycler(0)
                self.wake_up_inter_dashed_pen = et.utils.pg_pen_cycler(0, width=2.5, style="--")
            else:
                inter_color = et.utils.color_cycler(0)
                inter_color = f"{inter_color}50"
                self.wake_up_inter_dashed_pen = pg.mkPen(
                    inter_color, width=2.5, style=QtCore.Qt.DashLine
                )
                self.wake_up_inter_pen = pg.mkPen(inter_color, width=2)

    @staticmethod
    def create_sector_plot(
        title: str, num_sectors: int, start_m: float, zone_limits: npt.NDArray[np.float_]
    ) -> Tuple[pg.PlotItem, List[pg.QtWidgets.QGraphicsEllipseItem]]:
        sector_plot = pg.PlotItem(title=title)

        sector_plot.setAspectLocked()
        sector_plot.hideAxis("left")
        sector_plot.hideAxis("bottom")

        sectors = []
        limit_text = []

        range_html = (
            '<div style="text-align: center">'
            '<span style="color: #000000;font-size:12pt;">'
            "{}</span></div>"
        )

        if start_m == zone_limits[0]:
            x_offset = 0.7
        else:
            x_offset = 0

        pen = pg.mkPen("k", width=1)
        span_deg = 25
        for r in np.flip(np.arange(1, num_sectors + 2)):
            sector = pg.QtWidgets.QGraphicsEllipseItem(-r, -r, r * 2, r * 2)
            sector.setStartAngle(-16 * span_deg)
            sector.setSpanAngle(16 * span_deg * 2)
            sector.setPen(pen)
            sector_plot.addItem(sector)
            sectors.append(sector)

            if r != 1:
                limit = pg.TextItem(html=range_html, anchor=(0.5, 0.5), angle=25)
                x = r * np.cos(np.radians(span_deg))
                y = r * np.sin(np.radians(span_deg))
                limit.setPos(x - x_offset, y + 0.25)
                sector_plot.addItem(limit)
                limit_text.append(limit)

        sectors.reverse()

        if not start_m == zone_limits[0]:
            start_limit_text = pg.TextItem(html=range_html, anchor=(0.5, 0.5), angle=25)
            start_range_html = range_html.format(f"{start_m}")
            start_limit_text.setHtml(start_range_html)
            x = 1 * np.cos(np.radians(span_deg))
            y = 1 * np.sin(np.radians(span_deg))

            start_limit_text.setPos(x, y + 0.25)
            sector_plot.addItem(start_limit_text)

        unit_text = pg.TextItem(html=range_html, anchor=(0.5, 0.5))
        unit_html = range_html.format("[m]")
        unit_text.setHtml(unit_html)
        x = (num_sectors + 2) * np.cos(np.radians(span_deg))
        y = (num_sectors + 2) * np.sin(np.radians(span_deg))
        unit_text.setPos(x - x_offset, y + 0.25)
        sector_plot.addItem(unit_text)

        for (text_item, limit) in zip(limit_text, np.flip(zone_limits)):
            zone_range_html = range_html.format(np.around(limit, 1))
            text_item.setHtml(zone_range_html)

        return sector_plot, sectors

    def update(self, data: RefAppResult) -> None:
        if data.used_config == _Mode.NOMINAL_CONFIG:
            inter_threshold = self.nominal_config.inter_detection_threshold
            intra_threshold = self.nominal_config.intra_detection_threshold
            intra_pen = self.nominal_intra_pen
            intra_dashed_pen = self.nominal_intra_dashed_pen
            inter_pen = self.nominal_inter_pen
            inter_dashed_pen = self.nominal_inter_dashed_pen
        else:
            assert self.wake_up_config is not None
            inter_threshold = self.wake_up_config.inter_detection_threshold
            intra_threshold = self.wake_up_config.intra_detection_threshold
            intra_pen = self.wake_up_intra_pen
            intra_dashed_pen = self.wake_up_intra_dashed_pen
            inter_pen = self.wake_up_inter_pen
            inter_dashed_pen = self.wake_up_inter_dashed_pen

        self.time_fifo.append(data.service_result.tick_time)

        if data.switch_delay:
            self.intra_fifo.append(float("nan"))
            self.inter_fifo.append(float("nan"))
        else:
            self.intra_fifo.append(data.intra_presence_score)
            self.inter_fifo.append(data.inter_presence_score)

        while self.time_fifo[-1] - self.time_fifo[0] > self.history_length_s:
            self.time_fifo.pop(0)
            self.intra_fifo.pop(0)
            self.inter_fifo.pop(0)

        times = [t - self.time_fifo[-1] for t in self.time_fifo]

        # Intra presence

        if np.isnan(self.intra_fifo).all():
            m_hist = intra_threshold
        else:
            m_hist = np.maximum(float(np.nanmax(self.intra_fifo)), intra_threshold * 1.05)

        m_hist = self.intra_history_smooth_max.update(m_hist)

        self.intra_hist_plot.setYRange(0, m_hist)
        self.intra_hist_curve.setData(times, self.intra_fifo, connect="finite")
        self.intra_hist_curve.setPen(intra_pen)

        for line in self.intra_limit_lines:
            line.setPos(intra_threshold)
            line.setPen(intra_dashed_pen)

        # Inter presence

        if np.isnan(self.inter_fifo).all():
            m_hist = inter_threshold
        else:
            m_hist = np.maximum(float(np.nanmax(self.inter_fifo)), inter_threshold * 1.05)

        m_hist = self.inter_history_smooth_max.update(m_hist)

        self.inter_hist_plot.setYRange(0, m_hist)
        self.inter_hist_curve.setData(times, self.inter_fifo, connect="finite")
        self.inter_hist_curve.setPen(inter_pen)

        for line in self.inter_limit_lines:
            line.setPos(inter_threshold)
            line.setPen(inter_dashed_pen)

        # Sector

        brush = et.utils.pg_brush_cycler(7)
        for sector in self.nominal_sectors:
            sector.setBrush(brush)

        if not self.ref_app_config.wake_up_mode:
            sectors = self.nominal_sectors[1:]
            show_all_zones = self.show_all_detected_zones
            color_nominal = "white"
        else:
            if data.used_config == _Mode.WAKE_UP_CONFIG:
                sectors = self.wake_up_sectors[1:]
                show_all_zones = True
                color_wake_up = "#DFF1D6"
                color_nominal = "white"
            else:
                sectors = self.nominal_sectors[1:]
                show_all_zones = self.show_all_detected_zones
                color_wake_up = "white"
                color_nominal = "#DFF1D6"

            vb = self.nominal_sector_plot.getViewBox()
            vb.setBackgroundColor(color_nominal)
            vb = self.wake_up_sector_plot.getViewBox()
            vb.setBackgroundColor(color_wake_up)

            for sector in self.wake_up_sectors:
                sector.setBrush(brush)

        if data.presence_detected:
            self.color_zones(data, show_all_zones, sectors)
            self.switch_data = data
        elif data.switch_delay:
            self.color_zones(self.switch_data, True, self.wake_up_sectors[1:])

        self.nominal_sectors[0].setPen(pg.mkPen(color_nominal, width=1))
        self.nominal_sectors[0].setBrush(pg.mkBrush(color_nominal))

        if self.ref_app_config.wake_up_mode:
            self.wake_up_sectors[0].setPen(pg.mkPen(color_wake_up, width=1))
            self.wake_up_sectors[0].setBrush(pg.mkBrush(color_wake_up))

    @staticmethod
    def color_zones(
        data: RefAppResult,
        show_all_detected_zones: bool,
        sectors: List[pg.QtWidgets.QGraphicsEllipseItem],
    ) -> None:
        if show_all_detected_zones:
            for zone, (inter_value, intra_value) in enumerate(
                zip(data.inter_zone_detections, data.intra_zone_detections)
            ):
                if inter_value + intra_value == 2:
                    sectors[zone].setBrush(et.utils.pg_brush_cycler(2))
                elif inter_value == 1:
                    sectors[zone].setBrush(et.utils.pg_brush_cycler(0))
                elif intra_value == 1:
                    sectors[zone].setBrush(et.utils.pg_brush_cycler(1))
                elif data.used_config == _Mode.WAKE_UP_CONFIG:
                    assert data.wake_up_detections is not None
                    if data.wake_up_detections[zone] > 0:
                        sectors[zone].setBrush(pg.mkBrush("#b5afa0"))
        else:
            assert data.max_presence_zone is not None
            if data.max_presence_zone == data.max_intra_zone:
                sectors[data.max_presence_zone].setBrush(et.utils.pg_brush_cycler(1))
            else:
                sectors[data.max_presence_zone].setBrush(et.utils.pg_brush_cycler(0))


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

from typing import List, Optional

import attrs
import numpy as np
import numpy.typing as npt
from scipy.signal import filtfilt

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    ENVELOPE_FWHM_M,
    PERCEIVED_WAVELENGTH,
    AlgoProcessorConfigBase,
    ProcessorBase,
    get_distance_filter_coeffs,
    get_distance_offset,
    get_temperature_adjustment_factors,
)


TEMPERATURE_FILTER_CONSTANT = 0.95

# The same object can be seen at multiple subsweeps, objects close can be mereged.
MERGE_DISTANCE_M = 0.05
MERGE_SPEED_MPS = 0.05


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    num_std_treshold: float = attrs.field(default=5)
    num_mean_treshold: float = attrs.field(default=2)

    def _collect_validation_results(
        self, config: Optional[a121.SessionConfig]
    ) -> list[a121.ValidationResult]:
        return []


@attrs.frozen(kw_only=True)
class Target:
    distance: float = attrs.field(default=None)
    velocity: float = attrs.field(default=None)


@attrs.frozen(kw_only=True)
class SubsweepProcessorExtraResult:
    """
    Contains information for visualization in ET.
    """

    fft_map: npt.NDArray[np.float_] = attrs.field(default=None)
    fft_map_threshold: npt.NDArray[np.float_] = attrs.field(default=None)
    r: npt.NDArray[np.float_] = attrs.field(default=None)


@attrs.frozen(kw_only=True)
class SubsweepProcessorResult:
    targets: list[Target] = attrs.field(factory=list)
    extra_result: SubsweepProcessorExtraResult = attrs.field(factory=None)


@attrs.frozen(kw_only=True)
class ProcessorExtraResult:
    """
    Contains information for visualization in ET.
    """

    dv: float = attrs.field(default=None)  # velocity difference between fft-bins


@attrs.frozen(kw_only=True)
class ProcessorResult:
    targets: list[Target] = attrs.field(factory=list)
    time: float = attrs.field(default=None)
    extra_result: ProcessorExtraResult = attrs.field(factory=ProcessorExtraResult)
    subsweeps_extra_results: List[SubsweepProcessorExtraResult] = attrs.field(default=None)


@attrs.frozen(kw_only=True)
class SubsweepProcessorContext:
    sub_sweep_idx: int = attrs.field(default=None)


@attrs.frozen(kw_only=True)
class ProcessorContext:
    mean_sweeps: list[npt.NDArray[np.float_]] = attrs.field(factory=list)
    std_sweeps: list[npt.NDArray[np.float_]] = attrs.field(factory=list)
    reference_temperature: Optional[float] = attrs.field(default=None)
    loopback_peak_location_m: float = attrs.field(default=0.0)


class SubsweepProcessor:
    """Obstacle processor

    :param sensor_config: Sensor configuration
    :param metadata: Metadata yielded by the sensor config
    :param processor_config: Processor configuration
    """

    LOOPBACK_START_IDX = -48

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        processor_config: ProcessorConfig,
        proc_context: Optional[ProcessorContext] = None,
        ssproc_context: Optional[SubsweepProcessorContext] = None,
    ) -> None:
        if proc_context is None:
            self.proc_context = ProcessorContext()
        else:
            self.proc_context = proc_context

        if ssproc_context is None:
            self.ssproc_context = SubsweepProcessorContext()
        else:
            self.ssproc_context = ssproc_context

        self.sensor_config = sensor_config
        self.processor_config = processor_config

        # Threshold setting
        self.num_std_threshold = processor_config.num_std_treshold
        self.num_mean_threshold = processor_config.num_mean_treshold

        # Calbration result
        self.true_zero_dist_idx = 0

        self.offset_m = get_distance_offset(
            self.proc_context.loopback_peak_location_m, sensor_config.profile
        )

        if self.sensor_config.sweep_rate is None:
            self.sensor_config.sweep_rate = 100

        self.num_points = sensor_config.num_points
        self.dv = (
            PERCEIVED_WAVELENGTH
            / 2
            / self.sensor_config.sweeps_per_frame
            * self.sensor_config.sweep_rate
        )
        (self.b, self.a) = get_distance_filter_coeffs(
            sensor_config.profile, sensor_config.step_length
        )
        self.fwhm_points = ENVELOPE_FWHM_M[sensor_config.profile] / (
            APPROX_BASE_STEP_LENGTH_M * sensor_config.step_length
        )

        # Extra
        self.r = APPROX_BASE_STEP_LENGTH_M * (
            sensor_config.start_point + sensor_config.step_length * np.arange(self.num_points)
        )

    def process(
        self, subframe: npt.NDArray[np.complex_], temperature: float
    ) -> SubsweepProcessorResult:

        assert self.proc_context.reference_temperature is not None

        # Depth filtering of each sweep
        filtered_subframe = self.apply_depth_filter(subframe)

        # Range downsampling and fft in the sweep dimension
        fftframe = np.fft.fft(filtered_subframe, axis=0)
        abs_fftframe = np.abs(fftframe)
        abs_fftframe_extra = np.copy(abs_fftframe)  # Copy for plotting

        temp_diff = temperature - self.proc_context.reference_temperature
        sig_factor, noise_factor = get_temperature_adjustment_factors(
            temp_diff, self.sensor_config.profile
        )

        fft_map_threshold = np.tile(
            noise_factor
            * self.num_std_threshold
            * self.proc_context.std_sweeps[self.ssproc_context.sub_sweep_idx]
            * np.sqrt(self.sensor_config.sweeps_per_frame),
            (self.sensor_config.sweeps_per_frame, 1),
        )
        fft_map_threshold[0, :] = (
            fft_map_threshold[0, :]
            + sig_factor
            * self.num_mean_threshold
            * np.abs(self.proc_context.mean_sweeps[self.ssproc_context.sub_sweep_idx])
            * self.sensor_config.sweeps_per_frame
        )

        targets = []

        diff = abs_fftframe - fft_map_threshold
        spf = self.sensor_config.sweeps_per_frame
        while np.any(diff > 0):
            idx_max = np.unravel_index(np.argmax(diff), diff.shape)
            i_dist = get_interpolated_range_peak_index(
                diff[idx_max[0], :]
            )  # A non-flat threshold can move a peak slightly
            i_speed = get_interpolated_fft_peak_index(fftframe[:, idx_max[1]], int(idx_max[0]))

            distance = (
                APPROX_BASE_STEP_LENGTH_M
                * (self.sensor_config.start_point + self.sensor_config.step_length * i_dist)
                - self.offset_m
            )

            v = ((i_speed + spf / 2) % spf - spf / 2) * self.dv

            if 0 < idx_max[1] < (diff.shape[1] - 1):  # Disregard peaks at the limit of the range
                targets.append(Target(distance=distance, velocity=v))

            abs_fftframe = subtract_reflector_from_fftmap(
                abs_fftframe, int(idx_max[1]), int(idx_max[0]), int(self.fwhm_points)
            )
            diff = abs_fftframe - fft_map_threshold

        er = SubsweepProcessorExtraResult(
            fft_map=abs_fftframe_extra, fft_map_threshold=fft_map_threshold, r=self.r
        )

        return SubsweepProcessorResult(targets=targets, extra_result=er)

    def apply_depth_filter(self, frame: npt.NDArray[np.complex_]) -> npt.NDArray[np.complex_]:
        # Written as a separate function to be callable during detector calibration

        return np.array(filtfilt(self.b, self.a, frame), dtype=complex)

    def update_config(self, config: ProcessorConfig) -> None:
        pass


class Processor(ProcessorBase[ProcessorResult]):
    """Obstacle processor

    :param sensor_config: Sensor configuration
    :param processor_config: Processor configuration
    :param context: Optional processor context
    """

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        processor_config: ProcessorConfig,
        context: Optional[ProcessorContext] = None,
    ) -> None:
        if context is None:
            self.context = ProcessorContext()
        else:
            self.context = context

        self.sensor_config = sensor_config
        self.processor_config = processor_config

        self.num_subsweeps = sensor_config.num_subsweeps

        self.subsweep_processors: List[SubsweepProcessor] = []

        for i in range(self.num_subsweeps):
            subsweep_sensor_config_dict = sensor_config.to_dict()
            subsweep_sensor_config_dict["subsweeps"] = [sensor_config.subsweeps[i]]
            subsweep_sensor_config = a121.SensorConfig(**subsweep_sensor_config_dict)

            sspc = SubsweepProcessorContext(sub_sweep_idx=i)

            self.subsweep_processors.append(
                SubsweepProcessor(
                    sensor_config=subsweep_sensor_config,
                    processor_config=processor_config,
                    proc_context=self.context,
                    ssproc_context=sspc,
                )
            )

        # Calbration result
        self.true_zero_dist_idx = 0
        self.calbration_temperature = None

        self.filtered_sensor_temperature: Optional[float] = None
        self.temperature_filter_factor: float = 1 - np.exp(-TEMPERATURE_FILTER_CONSTANT)

        if self.sensor_config.sweep_rate is None:
            raise ValueError("The obstacle detector needs the sweep_rate set.")

        self.dv = (
            PERCEIVED_WAVELENGTH
            / 2
            / self.sensor_config.sweeps_per_frame
            * self.sensor_config.sweep_rate
        )

    def process(self, result: a121.Result) -> ProcessorResult:

        # Filter temperature
        if self.filtered_sensor_temperature is None:
            self.filtered_sensor_temperature = float(result.temperature)
        else:
            self.filtered_sensor_temperature *= self.temperature_filter_factor
            self.filtered_sensor_temperature += (
                1 - self.temperature_filter_factor
            ) * result.temperature

        subsweep_results = [
            proc.process(subframe, temperature=self.filtered_sensor_temperature)
            for subframe, proc in zip(result.subframes, self.subsweep_processors)
        ]

        merged_targets = self._merge_subsweep_targets(subsweep_results)
        subweeps_extra_results = [res.extra_result for res in subsweep_results]

        er = ProcessorExtraResult(dv=self.dv)

        return ProcessorResult(
            targets=merged_targets,
            time=result.tick_time,
            extra_result=er,
            subsweeps_extra_results=subweeps_extra_results,
        )

    def _merge_subsweep_targets(
        self, subsweep_results: List[SubsweepProcessorResult]
    ) -> List[Target]:

        # The same object can be seen at multiple subsweeps, objects close can be mereged.

        all_targets = [target for sr in subsweep_results for target in sr.targets]

        while True:
            closest_dist = 2.0  # Initialized to larger than one
            for i in range(len(all_targets)):
                for j in range(i):
                    d = (
                        (all_targets[i].velocity - all_targets[j].velocity) / MERGE_SPEED_MPS
                    ) ** 2
                    d += (
                        (all_targets[i].distance - all_targets[j].distance) / MERGE_DISTANCE_M
                    ) ** 2

                    if d < closest_dist:
                        closest_dist = d
                        ij = (i, j)

            if closest_dist < 1.0:
                t1 = all_targets[ij[0]]
                t2 = all_targets[ij[1]]
                all_targets.append(
                    Target(
                        distance=(t1.distance + t2.distance) / 2,
                        velocity=(t1.velocity + t2.velocity) / 2,
                    )
                )

                all_targets.remove(t1)
                all_targets.remove(t2)

            else:
                break

        return all_targets

    def apply_depth_filter(self, result: a121.Result) -> list[npt.NDArray[np.complex_]]:

        return [
            ssp.apply_depth_filter(r) for r, ssp in zip(result.subframes, self.subsweep_processors)
        ]

    def update_config(self, config: ProcessorConfig) -> None:
        pass


def apply_max_depth_filter(
    abs_sweep: npt.NDArray[np.float_], config: a121.SubsweepConfig
) -> npt.NDArray[np.float_]:
    """Filter sweep so that every point is max within +/- one fwhm"""

    filtered_sweep = np.zeros_like(abs_sweep)

    half_max_length_idx = int(
        ENVELOPE_FWHM_M[config.profile] / (APPROX_BASE_STEP_LENGTH_M * config.step_length)
    )

    for i, _ in enumerate(abs_sweep):
        filtered_sweep[i] = np.max(
            abs_sweep[
                max(0, i - half_max_length_idx) : min(i + half_max_length_idx, len(abs_sweep))
            ]
        )

    return filtered_sweep


def get_interpolated_range_peak_index(env: npt.NDArray[np.float_]) -> float:
    """Simple quadratic peak interpolation assuming equidistant points"""

    SMALL_NUMBER = 1e-10  # To avoid divide by zero for a flat peak.

    i = int(np.argmax(env))

    if i == 0 or i == len(env) - 1:
        return i

    return float(
        i
        + (env[i + 1] - env[i - 1]) / (4 * env[i] - 2 * env[i - 1] - 2 * env[i + 1] + SMALL_NUMBER)
    )


def get_interpolated_fft_peak_index(Y: npt.NDArray[np.complex_], idx: int) -> float:
    """FFT peak interpolation according to Quinn's method"""

    Y2 = np.abs(Y[idx]) ** 2
    idxp1 = (idx + 1) % len(Y)
    ap = (Y[idxp1].real * Y[idx].real + Y[idxp1].imag * Y[idx].imag) / Y2
    dp = -ap / (1.0 - ap)
    am = (Y[idx - 1].real * Y[idx].real + Y[idx - 1].imag * Y[idx].imag) / Y2
    dm = am / (1.0 - am)

    return float(idx + (dp if ((dp > 0) & (dm > 0)) else dm))


def subtract_reflector_from_fftmap(
    fftmap: npt.NDArray[np.float_], r_idx: int, f_idx: int, fwhm: float
) -> npt.NDArray[np.float_]:
    """Subtract signal from single reflector in fft map"""

    MARGIN_FACTOR = 2

    Nf, Nr = fftmap.shape
    map_range = np.clip(
        1 - np.abs(np.arange(Nr) - r_idx) / (MARGIN_FACTOR * fwhm), 0, np.Inf
    )  # Triangular envelope

    fs = np.arange(Nf)
    map_freq = np.sum([1 / (np.abs(f_idx + d * Nf - fs) + 1) for d in [-1, 0, 1]])

    peak = fftmap[f_idx, r_idx]

    return np.array(
        np.clip(
            fftmap - MARGIN_FACTOR * peak * np.outer(map_freq / np.max(map_freq), map_range),
            0,
            np.inf,
        )
    )



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import enum
from typing import Any, Dict, List, Optional, Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121, opser
from acconeer.exptool.a121._core.entities.configs.config_enums import PRF
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    PERCEIVED_WAVELENGTH,
    AlgoBase,
    AlgoConfigBase,
    calculate_loopback_peak_location,
)

from ._bilaterator import (
    Bilaterator,
    BilateratorResult,
)
from ._processors import (
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorResult,
    apply_max_depth_filter,
)


@attrs.mutable(kw_only=True)
class DetectorContext(AlgoBase):
    single_sensor_contexts: Dict[int, SingleSensorContext] = attrs.field(factory=dict)
    calibration_detector_config: Optional[DetectorConfig] = attrs.field(default=None)

    @property
    def sensor_ids(self) -> list[int]:
        return list(self.single_sensor_contexts.keys())


@attrs.mutable(kw_only=True)
class SingleSubsweepContext(AlgoBase):
    mean_sweep: npt.NDArray[np.float_]
    std_sweep: npt.NDArray[np.float_]


@attrs.mutable(kw_only=True)
class SingleSensorContext(AlgoBase):
    subsweep_contexts: List[SingleSubsweepContext] = attrs.field(factory=list)
    reference_temperature: Optional[float] = attrs.field(default=None)
    loopback_peak_location_m: Optional[float] = attrs.field(default=None)


@attrs.mutable(kw_only=True)
class DetectorConfig(AlgoConfigBase):
    start_m: float = attrs.field(default=0.15)
    """Start point of measurement interval in meters."""

    end_m: float = attrs.field(default=0.6)
    """End point of measurement interval in meters."""

    step_length: int = attrs.field(default=2)
    """Used to limit step length. """

    hwaas: int = attrs.field(default=12)
    """Hardware averaging. Higher gives better SNR but increase power
    consumption and lower sweep rate."""

    profile: a121.Profile = attrs.field(default=a121.Profile.PROFILE_3, converter=a121.Profile)
    """Profile, 1-5. Higher equals better SNR and lower increase resolution."""

    max_robot_speed: float = attrs.field(default=0.5)
    """Sets the sweep rate after the maximum robot speed in meters per second"""

    sweeps_per_frame: int = attrs.field(default=16)
    """Number of sweeps per frame."""

    num_frames_in_recorded_threshold: int = attrs.field(default=50)
    """Number of frames used when calibrating threshold."""

    num_std_threshold: float = attrs.field(default=5)
    """Number of standard deviations added to the threshold."""

    num_mean_threshold: float = attrs.field(default=2)
    """Number of means added to the threshold."""

    update_rate: float = attrs.field(default=50.0)
    """Sets the detector update rate."""

    subsweep_configurations: Optional[List[a121.SubsweepConfig]] = attrs.field(default=None)
    """Optional list of subsweep configurations that over-writes the sensor configuration."""

    enable_bilateration: bool = attrs.field(default=False)
    """Enable two-sensor bilateration."""

    bilateration_sensor_spacing_m: float = attrs.field(default=0.05)
    """The sensor spacing in meters."""

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if self.end_m < self.start_m:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "start_m",
                    "Must be smaller than 'Range end'",
                )
            )

            validation_results.append(
                a121.ValidationError(
                    self,
                    "end_m",
                    "Must be greater than 'Range start'",
                )
            )

        if self.sweeps_per_frame < 2:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "sweeps_per_frame",
                    "Must be larger than one for angle estimation.",
                )
            )

        return validation_results


@attrs.frozen(kw_only=True)
class DetectorResult:
    processor_results: Dict[int, ProcessorResult] = attrs.field(default=None)
    bilateration_result: Optional[BilateratorResult] = attrs.field(default=None)


@attrs.mutable(kw_only=True)
class DetectorStatus:
    detector_state: DetailedStatus
    ready_to_start: bool = attrs.field(default=False)


class DetailedStatus(enum.Enum):
    OK = enum.auto()
    END_LESSER_THAN_START = enum.auto()
    SENSOR_IDS_NOT_UNIQUE = enum.auto()
    CONTEXT_MISSING = enum.auto()
    CALIBRATION_MISSING = enum.auto()
    CONFIG_MISMATCH = enum.auto()


class Detector:
    """Obstacle detector
    :param client: Client
    :param sensor_id: Sensor id
    :param detector_config: Detector configuration
    :param context: Detector context
    """

    session_config: a121.SessionConfig
    processor_specs: List[ProcessorConfig]
    context: DetectorContext

    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_ids: list[int],
        detector_config: DetectorConfig,
        context: Optional[DetectorContext] = None,
    ) -> None:
        self.client = client
        self.sensor_ids = sensor_ids
        self.detector_config = detector_config

        self.started = False

        if context is None or not bool(context.single_sensor_contexts):
            self.context = DetectorContext(
                single_sensor_contexts={
                    sensor_id: SingleSensorContext() for sensor_id in self.sensor_ids
                }
            )
            self.detector_status = DetectorStatus(
                detector_state=DetailedStatus.CALIBRATION_MISSING,
                ready_to_start=False,
            )
        else:
            self.context = context
            self.detector_status = DetectorStatus(
                detector_state=DetailedStatus.OK,
                ready_to_start=True,
            )

        if self.detector_config.enable_bilateration:
            if len(sensor_ids) != 2:
                raise ValueError("Bilateration requires two, and only two, sensors.")
            self.bilaterator = Bilaterator(detector_config.bilateration_sensor_spacing_m)

        # Unless updated, assume that robot moves at max speed
        self.v_current = self.detector_config.max_robot_speed

        self.update_config(self.detector_config)
        self.true_zero_dist_idx = 0

    def _validate_ready_for_calibration(self) -> None:
        if self.started:
            raise RuntimeError("Already started")
        if self.processor_specs is None:
            raise ValueError("Processor specification not defined")
        if self.session_config is None:
            raise ValueError("Session config not defined")

    def calibrate_detector(self) -> None:
        """Run the required detector calibration routines, based on the detector config."""

        self._validate_ready_for_calibration()
        self._calibrate_offset()
        self._calibrate_threshold()

        self.context.calibration_detector_config = self.detector_config

        self.detector_status.detector_state = DetailedStatus.OK
        self.detector_status.ready_to_start = True

    def start(
        self,
        recorder: Optional[a121.Recorder] = None,
        _algo_group: Optional[h5py.Group] = None,
    ) -> None:
        if self.started:
            raise RuntimeError("Already started")

        if not self.detector_status.ready_to_start:
            raise RuntimeError("Not ready to start")

        sensor_config = self._get_sensor_config(self.detector_config)

        self.processors: Dict[int, Processor] = {}
        group = []
        for s_id in self.sensor_ids:
            sens_context = self.context.single_sensor_contexts[s_id]

            mean_sweeps: list[npt.NDArray[np.float_]] = []
            std_sweeps: list[npt.NDArray[np.float_]] = []

            for subsweep_idx in range(sensor_config.num_subsweeps):
                ssc = sens_context.subsweep_contexts[subsweep_idx]
                mean_sweeps.append(ssc.mean_sweep)
                std_sweeps.append(ssc.std_sweep)

            assert sens_context.loopback_peak_location_m is not None
            proc_context = ProcessorContext(
                mean_sweeps=mean_sweeps,
                std_sweeps=std_sweeps,
                loopback_peak_location_m=sens_context.loopback_peak_location_m,
                reference_temperature=sens_context.reference_temperature,
            )

            pc = ProcessorConfig(
                num_std_treshold=self.detector_config.num_std_threshold,
                num_mean_treshold=self.detector_config.num_mean_threshold,
            )

            self.processors[s_id] = Processor(
                sensor_config=sensor_config,
                processor_config=pc,
                context=proc_context,
            )

            group.append({s_id: sensor_config})

        self.session_config = a121.SessionConfig(group, extended=True)

        self.client.setup_session(self.session_config)

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                if _algo_group is None:
                    _algo_group = recorder.require_algo_group("obstacle_detector")

                _record_algo_data(
                    _algo_group,
                    self.sensor_ids,
                    self.detector_config,
                    self.context,
                )

            self.client.attach_recorder(recorder)

        self.client.start_session()

        self.detector_status.ready_to_start = False

        self.started = True

    def get_next(self) -> DetectorResult:
        if not self.started:
            raise RuntimeError("Not started")

        results = self.client.get_next()
        assert isinstance(results, list)

        processor_results: Dict[int, ProcessorResult] = {}

        for r in results:
            sens_id = list(r.keys())[0]  # get the only sensor-id in that dict.
            processor_results[sens_id] = self.processors[sens_id].process(r[sens_id])

        if self.detector_config.enable_bilateration:
            target_list_1 = processor_results[self.sensor_ids[0]].targets
            target_list_2 = processor_results[self.sensor_ids[1]].targets
            bilateration_result = self.bilaterator.process(target_list_1, target_list_2)
        else:
            bilateration_result = None

        return DetectorResult(
            processor_results=processor_results, bilateration_result=bilateration_result
        )

    def update_robot_speed(self, v_current: float) -> None:
        self.v_current = v_current

    def get_obstacle_angle(self, v_measured: float) -> float:
        """Convert the measured speed of an object to an angle"""
        return float(np.arccos(v_measured / self.v_current))

    def stop(self) -> Any:
        """Stops the measurement session."""
        if not self.started:
            raise RuntimeError("Already stopped")

        self.client.stop_session()

        self.detector_status.ready_to_start = True

        recorder = self.client.detach_recorder()
        if recorder is None:
            recorder_result = None
        else:
            recorder_result = recorder.close()

        self.started = False

        return recorder_result

    def update_config(self, config: DetectorConfig) -> None:
        """Updates the session config and processor specification based on the detector
        configuration."""
        (
            self.session_config,
            self.processor_specs,
        ) = self._detector_to_session_config_and_processor_specs(
            config=config, sensor_ids=self.sensor_ids
        )

        if not (self.context.calibration_detector_config == config):
            self.detector_status.detector_state = DetailedStatus.CALIBRATION_MISSING
            self.detector_status.ready_to_start = False

    @classmethod
    def _get_sensor_config(cls, detector_config: DetectorConfig) -> a121.SensorConfig:
        if detector_config.subsweep_configurations is None:
            start_p = int(detector_config.start_m / APPROX_BASE_STEP_LENGTH_M)
            num_p = int(
                (detector_config.end_m - detector_config.start_m)
                / (APPROX_BASE_STEP_LENGTH_M * detector_config.step_length)
            )

            subsweeps = [
                a121.SubsweepConfig(
                    start_point=start_p,
                    num_points=num_p,
                    step_length=detector_config.step_length,
                    profile=detector_config.profile,
                    hwaas=detector_config.hwaas,
                    prf=PRF.PRF_15_6_MHz,
                )
            ]

        else:
            subsweeps = detector_config.subsweep_configurations

        for subsweep in subsweeps:
            subsweep.phase_enhancement = True

        return a121.SensorConfig(
            sweeps_per_frame=detector_config.sweeps_per_frame,
            sweep_rate=detector_config.max_robot_speed / (PERCEIVED_WAVELENGTH / 2),
            inter_sweep_idle_state=a121.IdleState.READY,
            inter_frame_idle_state=a121.IdleState.READY,
            subsweeps=subsweeps,
        )

    @classmethod
    def _detector_to_session_config_and_processor_specs(
        cls, config: DetectorConfig, sensor_ids: list[int]
    ) -> Tuple[a121.SessionConfig, list[ProcessorConfig]]:
        processor_spec = ProcessorConfig(
            num_std_treshold=config.num_std_threshold,
            num_mean_treshold=config.num_mean_threshold,
        )

        sensor_config = cls._get_sensor_config(config)
        return (
            a121.SessionConfig({s_id: sensor_config for s_id in sensor_ids}, extended=True),
            [processor_spec] * len(sensor_ids),
        )

    def _calibrate_offset(self) -> None:
        """Estimates sensor offset error based on loopback measurement."""

        self._validate_ready_for_calibration()

        sensor_config = a121.SensorConfig(
            start_point=-30,
            num_points=50,
            step_length=1,
            profile=a121.Profile.PROFILE_1,
            hwaas=64,
            sweeps_per_frame=1,
            enable_loopback=True,
            phase_enhancement=True,
        )

        # Sensors running simultaneously, profile 1 and direct leakage give low risk of cross talk.
        session_config = a121.SessionConfig(
            {sensor_id: sensor_config for sensor_id in self.sensor_ids},
            extended=True,
        )

        self.client.setup_session(session_config)
        self.client.start_session()
        extended_result = self.client.get_next()
        self.client.stop_session()

        assert isinstance(extended_result, list)

        for sensor_id, context in self.context.single_sensor_contexts.items():
            context.loopback_peak_location_m = calculate_loopback_peak_location(
                extended_result[0][sensor_id], sensor_config
            )

    def _calibrate_threshold(self) -> None:
        self._validate_ready_for_calibration()

        sensor_config = self._get_sensor_config(self.detector_config)
        session_config = a121.SessionConfig(
            {sensor_id: sensor_config for sensor_id in self.sensor_ids}, extended=True
        )
        self.client.setup_session(session_config)

        result_list: list[dict[int, a121.Result]] = []
        self.client.start_session()
        for _ in range(self.detector_config.num_frames_in_recorded_threshold):
            extended_result = self.client.get_next()
            assert isinstance(extended_result, list)
            assert all(isinstance(r, dict) for r in extended_result)
            result_list.append({i: r for d in extended_result for i, r in d.items()})
        self.client.stop_session()

        # Processor used for applying the depth filter
        processor = Processor(sensor_config=sensor_config, processor_config=ProcessorConfig())

        for s_id in self.sensor_ids:
            results: list[a121.Result] = [r[s_id] for r in result_list]
            depth_filtered_data = [processor.apply_depth_filter(result) for result in results]
            threshold_temp = np.mean([r.temperature for r in results])

            self.context.single_sensor_contexts[s_id].subsweep_contexts = []
            for subsweep_idx in range(sensor_config.num_subsweeps):
                subsweep_filtered_data = np.array(
                    [dfd[subsweep_idx] for dfd in depth_filtered_data]
                )

                abs_mean = np.abs(
                    np.mean(subsweep_filtered_data, axis=(0, 1))
                )  # Over the frame- and sweep dimension
                std = np.std(subsweep_filtered_data, axis=(0, 1))

                subsweep_mean = apply_max_depth_filter(
                    abs_mean, sensor_config.subsweeps[subsweep_idx]
                )

                ssc = SingleSubsweepContext(mean_sweep=subsweep_mean, std_sweep=std)
                self.context.single_sensor_contexts[s_id].subsweep_contexts.append(ssc)

            self.context.single_sensor_contexts[s_id].reference_temperature = threshold_temp

    @classmethod
    def get_detector_status(
        cls,
        config: DetectorConfig,
        context: DetectorContext,
        sensor_ids: list[int],
    ) -> DetectorStatus:
        """Returns the detector status along with the detector state."""

        if config.end_m < config.start_m:
            return DetectorStatus(
                detector_state=DetailedStatus.END_LESSER_THAN_START,
                ready_to_start=False,
            )

        if len(sensor_ids) != len(set(sensor_ids)):
            return DetectorStatus(
                detector_state=DetailedStatus.SENSOR_IDS_NOT_UNIQUE,
                ready_to_start=False,
            )

        if context.single_sensor_contexts is None:
            return DetectorStatus(
                detector_state=DetailedStatus.CONTEXT_MISSING,
                ready_to_start=False,
            )

        # Offset calibration is always performed as a part of the detector calibration process.
        # Use this as indication whether detector calibration has been performed.
        calibration_missing = np.any(
            [
                ctx.loopback_peak_location_m is None
                for ctx in context.single_sensor_contexts.values()
            ]
        )
        config_mismatch = context.calibration_detector_config != config

        if calibration_missing:
            detector_state = DetailedStatus.CALIBRATION_MISSING
        elif config_mismatch:
            detector_state = DetailedStatus.CONFIG_MISMATCH
        else:
            detector_state = DetailedStatus.OK

        return DetectorStatus(
            detector_state=detector_state,
            ready_to_start=(detector_state == DetailedStatus.OK),
        )


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_ids: list[int],
    config: DetectorConfig,
    context: DetectorContext,
) -> None:
    algo_group.create_dataset(
        "sensor_ids",
        data=sensor_ids,
        track_times=False,
    )
    _create_h5_string_dataset(algo_group, "detector_config", config.to_json())

    context_group = algo_group.create_group("context")
    opser.serialize(context, context_group)


def _load_algo_data(
    algo_group: h5py.Group,
) -> Tuple[int, DetectorConfig, DetectorContext]:
    sensor_id = algo_group["sensor_ids"][()]
    config = DetectorConfig.from_json(algo_group["detector_config"][()])

    context_group = algo_group["context"]
    context = opser.deserialize(context_group, DetectorContext)

    return sensor_id, config, context



# Copyright (c) Acconeer AB, 2023
# All rights reserved


import h5py
import matplotlib.pyplot as plt

from acconeer.exptool.a121 import H5Record, _ReplayingClient, _StopReplay
from acconeer.exptool.a121.algo.distance import Detector, DetectorConfig, DetectorContext


FILEPATH = ""


def main():
    with h5py.File(FILEPATH, "r") as file:
        algo_group = file["algo"]

        sensor_ids = algo_group["sensor_ids"][()].tolist()
        detector_config = DetectorConfig.from_json(algo_group["detector_config"][()])
        context_group = algo_group["context"]
        context = DetectorContext.from_h5(context_group)

        record = H5Record(file)
        client = _ReplayingClient(record, realtime_replay=False)

        detector = Detector(
            client=client,
            sensor_ids=sensor_ids,
            detector_config=detector_config,
            context=context,
        )

        detector.start()

        fig, axs = plt.subplots(ncols=2, sharex=True)
        fig.set_figwidth(15)

        while True:
            try:
                result = detector.get_next()

                for processor_result in result[sensor_ids[0]].processor_results:
                    axs[0].plot(
                        processor_result.extra_result.distances_m,
                        processor_result.extra_result.abs_sweep,
                        "b",
                    )
                    axs[0].plot(
                        processor_result.extra_result.distances_m,
                        processor_result.extra_result.used_threshold,
                        "r",
                    )

                axs[1].plot(result[sensor_ids[0]].distances, result[sensor_ids[0]].strengths, "k.")
            except _StopReplay:
                break

    axs[0].legend(["Sweep", "Threshold"], loc="upper left")

    axs[0].set_xlabel("Distance (m)")
    axs[0].set_ylabel("Amplitude")
    axs[0].grid()

    axs[1].set_xlabel("Distance (m)")
    axs[1].set_ylabel("Strength")
    axs[1].grid()

    plt.legend()
    plt.suptitle(FILEPATH)
    plt.show()


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

from typing import List, Optional

import attrs

import acconeer.exptool as et
from acconeer.exptool._links import BufferedLink, ExploreSerialLink, NullLink, SocketLink, USBLink
from acconeer.exptool.a121._core.entities import (
    ClientInfo,
    SensorCalibration,
    SerialInfo,
    SessionConfig,
    USBInfo,
)
from acconeer.exptool.a121._core.utils import iterate_extended_structure
from acconeer.exptool.utils import SerialDevice, USBDevice

from .common_client import ClientError


def get_one_serial_device() -> SerialDevice:
    acconeer_serial_devices: List[SerialDevice] = [
        device for device in et.utils.get_serial_devices() if device.name is not None
    ]

    if not acconeer_serial_devices:
        raise ClientError("No serial devices detected. Cannot auto detect.")
    elif len(acconeer_serial_devices) > 1:
        devices_string = "".join([f" - {dev}\n" for dev in acconeer_serial_devices])
        raise ClientError("There are multiple devices detected. Specify one:\n" + devices_string)
    else:
        return acconeer_serial_devices[0]


def get_one_usb_device(only_accessible: bool = False) -> USBDevice:
    usb_devices = et.utils.get_usb_devices(only_accessible=only_accessible)
    if not usb_devices:
        raise ClientError("No USB devices detected. Cannot auto detect.")
    elif len(usb_devices) > 1:
        devices_string = "".join([f" - {dev}\n" for dev in usb_devices])
        raise ClientError("There are multiple devices detected. Specify one:\n" + devices_string)
    else:
        return usb_devices[0]


def link_factory(client_info: ClientInfo) -> BufferedLink:

    if client_info.socket is not None:
        return SocketLink(host=client_info.socket.ip_address, port=client_info.socket.tcp_port)

    if client_info.serial is not None:
        return ExploreSerialLink(
            port=client_info.serial.port,
        )

    if client_info.usb is not None:
        usb_device = None
        if client_info.usb.serial_number is not None:
            usb_device = et.utils.get_usb_device_by_serial(
                client_info.usb.serial_number, only_accessible=False
            )
        else:
            usb_device = get_one_usb_device()

        if usb_device is not None:
            return USBLink(
                vid=usb_device.vid,
                pid=usb_device.pid,
                serial=usb_device.serial,
            )

    return NullLink()


def autodetermine_client_link(client_info: ClientInfo) -> ClientInfo:
    error_message = ""
    try:
        usb_info = client_info.usb
        if usb_info is None:
            usb_device = get_one_usb_device(only_accessible=True)
            usb_info = USBInfo(
                vid=usb_device.vid, pid=usb_device.pid, serial_number=usb_device.serial
            )
        return attrs.evolve(
            client_info,
            usb=usb_info,
        )

    except ClientError as exc:
        error_message += f"\nUSB: {str(exc)}"
        pass

    try:
        serial_info = client_info.serial
        if serial_info is None:
            serial_info = SerialInfo(port=str(get_one_serial_device().port))
        return attrs.evolve(client_info, serial=serial_info)
    except ClientError as exc:
        error_message += f"\nSerial: {str(exc)}"

    raise ClientError(f"Cannot auto detect:{error_message}")


def get_calibrations_provided(
    session_config: SessionConfig,
    calibrations: Optional[dict[int, SensorCalibration]] = None,
) -> dict[int, bool]:
    calibrations_provided = {}
    for _, sensor_id, _ in iterate_extended_structure(session_config.groups):
        if calibrations:
            calibrations_provided[sensor_id] = sensor_id in calibrations
        else:
            calibrations_provided[sensor_id] = False

    return calibrations_provided



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from acconeer.exptool import a121


# Client creation
client = a121.Client.open(ip_address="192.168.0.1")

# Session setup, just like the other examples.
config = a121.SessionConfig()
client.setup_session(config)


# Here we specify the file name for the H5Recorder.
# The H5Recorder is an object that saves frames directly to a H5-file.
filename = "data.h5"

# Here we create and attach a H5Recorder to the Client.
# The H5Recorder will sample all frames retrieved in the client.
# Once the "with"-block have been exited, the H5Recorder will
# wrap up and close the file.
with a121.H5Recorder(filename, client):
    client.start_session()

    n = 10
    for i in range(n):
        # Client will send its Results to the H5Recorder.
        client.get_next()
        print(f"Result {i + 1}/{n} was sampled")

    client.stop_session()

client.close()

with a121.open_record(filename) as record:
    print(record)



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import attrs
import numpy as np

from acconeer.exptool.a121.algo.obstacle._processors import (
    Target,
)


@attrs.frozen(kw_only=True)
class BilateratorResult:
    beta_degs: list[float] = attrs.field()  # Bilateration angle
    velocities_m_s: list[float] = attrs.field()
    distances_m: list[float] = attrs.field()


DISTANCE_LIMIT_FACTOR = 1.2


class Bilaterator:
    """Bilateration class that aggregates the output from two obstacle detectors

    Aggregates result, based on selected peak sorting strategy, from underlying Processors objects.
    """

    def __init__(self, sensor_seperation: float):
        self.sensor_seperation = sensor_seperation

    def process(
        self, target_list_1: list[Target], target_list_2: list[Target]
    ) -> BilateratorResult:

        if not target_list_1 or not target_list_2:
            return BilateratorResult(beta_degs=[], velocities_m_s=[], distances_m=[])

        dist_diff = target_list_1[0].distance - target_list_2[0].distance
        dist_mean = (target_list_1[0].distance + target_list_2[0].distance) / 2
        velocity_mean = (target_list_1[0].velocity + target_list_2[0].velocity) / 2

        if np.abs(dist_diff) > DISTANCE_LIMIT_FACTOR * self.sensor_seperation:
            # Unreasonable distance difference
            return BilateratorResult(beta_degs=[], velocities_m_s=[], distances_m=[])

        if dist_diff > self.sensor_seperation:
            beta_deg = 90
        elif dist_diff < -self.sensor_seperation:
            beta_deg = -90
        else:
            beta_deg = 180 / np.pi * np.arcsin(dist_diff / self.sensor_seperation)

        return BilateratorResult(
            beta_degs=[beta_deg], velocities_m_s=[velocity_mean], distances_m=[dist_mean]
        )



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from . import pidgets
from .attrs_config_editor import AttrsConfigEditor
from .collapsible_widget import CollapsibleWidget
from .data_editor import DataEditor
from .goto_resource_tab_button import GotoResourceTabButton
from .json_save_load_buttons import PresentationType, PresenterFunc
from .metadata_view import ExtendedMetadataView, MetadataView, SmartMetadataView
from .misc_error_view import MiscErrorView
from .perf_calc_view import ExtendedPerfCalcView, PerfCalcView, SmartPerfCalcView
from .sensor_config_editor import SensorConfigEditor
from .session_config_editor import SessionConfigEditor
from .subsweep_config_editor import SubsweepConfigEditor
from .two_sensor_ids_editor import TwoSensorIdsEditor
from .types import PidgetFactoryMapping, PidgetGroupFactoryMapping
from .utils import GroupBox



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import copy
from functools import partial
from typing import Any, Optional, Sequence, Type, TypeVar, Union, cast

import attrs

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QVBoxLayout, QWidget

from acconeer.exptool import a121
from acconeer.exptool.a121._core import Criticality

from .data_editor import DataEditor
from .json_save_load_buttons import (
    JsonPresentable,
    PresenterFunc,
    create_json_save_load_buttons_from_type,
)
from .pidgets import FlatPidgetGroup, Pidget, PidgetGroup, PidgetHook, WidgetHook
from .types import PidgetFactoryMapping, PidgetGroupFactoryMapping
from .utils import GroupBox


T = TypeVar("T", bound=JsonPresentable)


def _to_group_factory_mapping(
    factory_mapping: Union[PidgetFactoryMapping, PidgetGroupFactoryMapping]
) -> PidgetGroupFactoryMapping:
    if factory_mapping == {}:
        return {}

    (first_key, *_) = factory_mapping

    # The casts boils down to "non-transferable" type narrowing
    # of a Mapping (typically a dict) given its keys.
    #   In this function, we want to do type narrowing of the parameter type
    # (rewritten as dict[Union[str, PidgetGroup], ...]). The key will have type
    # Union[str, PidgetGroup], which can be narrowed to str or PidgetGroup.
    # This type narrowing is not transferred to the original dict, requiring a cast.
    if isinstance(first_key, PidgetGroup):
        return cast(PidgetGroupFactoryMapping, factory_mapping)

    if isinstance(first_key, str):
        return {FlatPidgetGroup(): cast(PidgetFactoryMapping, factory_mapping)}

    raise RuntimeError(
        "factory_mapping was neither a PidgetFactoryMappingi nor a PidgetGroupFactoryMapping"
    )


class AttrsConfigEditor(DataEditor[Optional[T]]):
    _config: Optional[T]
    _erroneous_aspects: set[str]

    sig_update = Signal(object)

    def __init__(
        self,
        title: str,
        factory_mapping: Union[PidgetFactoryMapping, PidgetGroupFactoryMapping],
        config_type: Type[T],
        *,
        save_load_buttons: bool = True,
        extra_presenter: PresenterFunc = lambda i, t: None,
        parent: Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent=parent)
        self._config = None
        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(11)

        group_box = GroupBox.vertical(
            title,
            right_header=(
                create_json_save_load_buttons_from_type(
                    self, config_type, extra_presenter=extra_presenter
                )
                if save_load_buttons
                else None
            ),
            parent=self,
        )
        self.layout().addWidget(group_box)

        self._pidget_mapping: dict[str, Pidget] = {}
        self._pidget_hooks: dict[str, Sequence[PidgetHook]] = {}
        self._group_widgets: list[QWidget] = []
        self._group_hooks: list[Sequence[WidgetHook]] = []
        self._erroneous_aspects = set()

        for pidget_group, factory_mapping in _to_group_factory_mapping(factory_mapping).items():
            pidgets = []
            for aspect, factory in factory_mapping.items():
                pidget = factory.create(group_box)
                pidget.sig_update.connect(partial(self._update_config_aspect, aspect))

                self._pidget_mapping[aspect] = pidget
                self._pidget_hooks[aspect] = factory.hooks

                pidgets.append(pidget)

            group_widget = pidget_group.get_container(pidgets)
            group_box.layout().addWidget(group_widget)

            self._group_widgets.append(group_widget)
            self._group_hooks.append(pidget_group.hooks)

    def set_data(self, config: Optional[T]) -> None:
        if self._config == config:
            return

        self._config = config
        self.setEnabled(config is not None)

        if self._config is None:
            return

        for aspect, pidget in self._pidget_mapping.items():
            if aspect in self._erroneous_aspects:
                continue
            config_value = getattr(self._config, aspect)
            pidget.set_data(config_value)

        for aspect, hooks in self._pidget_hooks.items():
            for hook in hooks:
                hook(self._pidget_mapping[aspect], self._pidget_mapping)

        for group_widget, hooks in zip(self._group_widgets, self._group_hooks):
            for hook in hooks:
                hook(group_widget, self._pidget_mapping)

    def get_data(self) -> Optional[T]:
        return copy.deepcopy(self._config)

    def handle_validation_results(
        self, results: list[a121.ValidationResult]
    ) -> list[a121.ValidationResult]:
        for aspect, pidget in self._pidget_mapping.items():
            if aspect not in self._erroneous_aspects:
                pidget.set_note_text("")

        unhandled_results: list[a121.ValidationResult] = []

        for result in results:
            if not self._handle_validation_result(result):
                unhandled_results.append(result)

        return unhandled_results

    @property
    def is_ready(self) -> bool:
        return self._erroneous_aspects == set()

    def _handle_validation_result(self, result: a121.ValidationResult) -> bool:
        if result.aspect is None:
            return False

        if result.aspect in self._erroneous_aspects:
            # If there is an erroneous aspect in the GUI, we do not want to overwrite that.
            # Once the erroneous aspect is handled with, the same validation result will
            # come through here anyway
            return True

        result_handled = False

        if result.source == self._config:
            for aspect, pidget in self._pidget_mapping.items():
                if result.aspect == aspect:
                    self._pidget_mapping[result.aspect].set_note_text(
                        result.message, result.criticality
                    )
                    result_handled = True

        return result_handled

    def _update_config_aspect(self, aspect: str, value: Any) -> None:
        if self._config is None:
            return

        try:
            # The passing of "self._config: T" is not type safe. We cannot know
            # if "T" is an attrs class in the scope of this class without binding
            # "T" to a common superclass.
            # "AlgoConfigBase" is a candidate, but that is part of the algo-package.
            config = attrs.evolve(self._config, **{aspect: value})  # type: ignore[misc]
        except Exception as e:
            self._erroneous_aspects.add(aspect)
            self._pidget_mapping[aspect].set_note_text(e.args[0], Criticality.ERROR)

            # this emit needs to be done to signal that "is_ready" has changed.
            self.sig_update.emit(self.get_data())
        else:
            self._pidget_mapping[aspect].set_note_text(None)
            self._erroneous_aspects.discard(aspect)

            self.set_data(config)
            self.sig_update.emit(config)



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from ._links import (
    BufferedLink,
    ExploreSerialLink,
    NullLink,
    NullLinkError,
    SerialLink,
    SerialProcessLink,
    SocketLink,
    USBLink,
)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

import logging
import multiprocessing as mp
import platform
import queue
import signal
import socket
import threading as th
import traceback
from time import sleep, time
from typing import Any, Optional, Tuple, Union

import serial
from typing_extensions import Protocol

from acconeer.exptool._pyusb.pyusbcomm import PyUsbCdc


ComPort: Any
try:
    from acconeer.exptool._winusbcdc.usb_cdc import ComPort
except ImportError:
    ComPort = None

log = logging.getLogger(__name__)


class LinkError(RuntimeError):
    pass


class NullLinkError(RuntimeError):
    pass


class BufferedLink(Protocol):
    def connect(self) -> None:
        """Establishes a connection."""
        ...

    @property
    def timeout(self) -> float:
        """Return link timout."""
        ...

    @timeout.setter
    def timeout(self, timeout: float) -> None:
        """Set return link timeout."""
        ...

    def recv(self, num_bytes: int) -> bytes:
        """Recieves `num_bytes` bytes."""
        ...

    def send(self, bytes_: bytes) -> None:
        """Sends all `bytes_` over the link."""
        ...

    def disconnect(self) -> None:
        """Tears down the connection."""
        ...

    def recv_until(self, byte_sequence: bytes) -> bytes:
        """Collects all bytes until `byte_sequence` is encountered,
        returning what was collected
        """
        ...


class NullLink(BufferedLink):
    ERROR = NullLinkError("Link is undetermined.")
    """Link null object.

    :raises: ``RuntimeError`` if any of its methods is called.
    """

    def connect(self) -> None:
        raise self.ERROR

    @property
    def timeout(self) -> float:
        raise self.ERROR

    @timeout.setter
    def timeout(self, timeout: float) -> None:
        raise self.ERROR

    def recv(self, num_bytes: int) -> bytes:
        raise self.ERROR

    def send(self, bytes_: bytes) -> None:
        raise self.ERROR

    def disconnect(self) -> None:
        raise self.ERROR

    def recv_until(self, byte_sequence: bytes) -> bytes:
        raise self.ERROR


class BaseLink(BufferedLink):
    DEFAULT_TIMEOUT = 2.0

    def __init__(self) -> None:
        self._timeout = self.DEFAULT_TIMEOUT

    @property
    def timeout(self) -> float:
        return self._timeout

    @timeout.setter
    def timeout(self, val: float) -> None:
        self._timeout = val
        self._update_timeout()

    def _update_timeout(self) -> None:
        pass


class SocketLink(BaseLink):
    _CHUNK_SIZE = 4096
    _PORT = 6110

    def __init__(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        super().__init__()
        self._host = host
        self._sock: Optional[socket.socket] = None
        self._buf: bytearray = bytearray()
        self._port: int = self._PORT if (port is None) else port

    def _update_timeout(self) -> None:
        if self._sock is not None:
            self._sock.settimeout(self._timeout)

    def connect(self) -> None:
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._update_timeout()

        try:
            self._sock.connect((self._host, self._port))
        except OSError as e:
            self._sock.close()
            self._sock = None
            raise LinkError("failed to connect") from e

        self._buf = bytearray()

    def recv(self, num_bytes: int) -> bytes:
        assert self._sock is not None
        while len(self._buf) < num_bytes:
            try:
                r = self._sock.recv(self._CHUNK_SIZE)
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        data = self._buf[:num_bytes]
        self._buf = self._buf[num_bytes:]
        return data

    def recv_until(self, bs: bytes) -> bytes:
        assert self._sock is not None
        t0 = time()
        while True:
            try:
                i = self._buf.index(bs)
            except ValueError:
                pass
            else:
                break

            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

            try:
                r = self._sock.recv(self._CHUNK_SIZE)
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        i += 1
        data = self._buf[:i]
        self._buf = self._buf[i:]

        return data

    def send(self, data: bytes) -> None:
        assert self._sock is not None
        self._sock.sendall(data)

    def disconnect(self) -> None:
        if self._sock is not None:
            self._sock.shutdown(socket.SHUT_RDWR)
            self._sock.close()
            self._sock = None
        self._buf = bytearray()


class BaseSerialLink(BaseLink):
    def __init__(self, baudrate: int = 115200) -> None:
        super().__init__()
        self._timeout = self.DEFAULT_TIMEOUT
        self._baudrate = baudrate


class SerialLink(BaseSerialLink):
    def __init__(self, port: Optional[str] = None, flowcontrol: bool = False) -> None:
        super().__init__()
        self._port = port
        self._ser: Optional[serial.Serial] = None
        self._flowcontrol = flowcontrol

    def _update_timeout(self) -> None:
        if self._ser is not None:
            self._ser.timeout = self._timeout

    def send_break(self) -> None:
        assert self._ser is not None
        self._ser.send_break()
        sleep(1.0)
        self._ser.flushInput()

    def connect(self) -> None:
        self._ser = serial.Serial()
        self._ser.baudrate = self._baudrate
        self._ser.port = self._port
        self._ser.rtscts = self._flowcontrol
        self._update_timeout()
        self._ser.open()

        if platform.system().lower() == "windows":
            self._ser.set_buffer_size(rx_size=10**6, tx_size=10**6)

    def recv(self, num_bytes: int) -> bytes:
        assert self._ser is not None
        data = bytearray(self._ser.read(num_bytes))

        if not len(data) == num_bytes:
            raise LinkError("recv timeout")

        return data

    def recv_until(self, bs: bytes) -> bytes:
        assert self._ser is not None
        data = bytearray(self._ser.read_until(bs))

        if bs not in data:
            raise LinkError("recv timeout")

        return data

    def send(self, data: bytes) -> None:
        assert self._ser is not None
        self._ser.write(data)

    def disconnect(self) -> None:
        if self._ser is not None:
            self._ser.close()
            self._ser = None

    @property
    def baudrate(self) -> int:
        return self._baudrate

    @baudrate.setter
    def baudrate(self, new_baudrate: int) -> None:
        self._baudrate = new_baudrate

        if self._ser is not None and self._ser.is_open:
            self._ser.baudrate = new_baudrate


class ExploreSerialLink(SerialLink):
    _SERIAL_READ_PACKET_SIZE = 65536
    _SERIAL_PACKET_TIMEOUT = 0.01
    _SERIAL_WRITE_TIMEOUT = 1.0

    def __init__(self, port: str, flowcontrol: bool = True) -> None:
        super().__init__(port, flowcontrol)
        self._buf = bytearray()

    def _update_timeout(self) -> None:
        pass

    def connect(self) -> None:
        self._ser = serial.Serial(
            timeout=self._SERIAL_PACKET_TIMEOUT,
            write_timeout=self._SERIAL_WRITE_TIMEOUT,
            exclusive=True,
        )
        self._ser.baudrate = self._baudrate
        self._ser.port = self._port
        self._ser.rtscts = self._flowcontrol
        self._ser.open()
        self._buf = bytearray()

        if platform.system().lower() == "windows":
            self._ser.set_buffer_size(rx_size=10**6, tx_size=10**6)

        self.send_break()

    def recv(self, num_bytes: int) -> bytes:
        assert self._ser is not None
        t0 = time()
        while len(self._buf) < num_bytes:
            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

            try:
                r = bytearray(self._ser.read(self._SERIAL_READ_PACKET_SIZE))
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        data = self._buf[:num_bytes]
        self._buf = self._buf[num_bytes:]
        return data

    def recv_until(self, bs: bytes) -> bytes:
        assert self._ser is not None
        t0 = time()
        while True:
            try:
                i = self._buf.index(bs)
            except ValueError:
                pass
            else:
                break

            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

            try:
                r = bytearray(self._ser.read(self._SERIAL_READ_PACKET_SIZE))
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        i += 1
        data = self._buf[:i]
        self._buf = self._buf[i:]

        return data


class USBLink(BaseLink):
    def __init__(
        self, vid: Optional[int] = None, pid: Optional[int] = None, serial: Optional[str] = None
    ) -> None:
        super().__init__()
        self._vid = vid
        self._pid = pid
        self._serial = serial

    def connect(self) -> None:
        # First try 'ComPort', will be set if platorm == windows
        port_type = ComPort
        if ComPort is None:
            # Fallback to  'PyUsbCdc', will be used if platform != windows
            port_type = PyUsbCdc

        if self._vid and self._pid:
            self._port = port_type(vid=self._vid, pid=self._pid, serial=self._serial, start=False)
        else:
            raise LinkError("Must have vid and pid for usb device")

        if not self._port.open():
            raise LinkError(f"Unable to connect to port (vid={self._vid}, pid={self._pid}")

        self._buf = bytearray()
        self.send_break()

    def send_break(self) -> None:
        self._port.send_break()
        sleep(1.0)
        self._port.reset_input_buffer()

    def recv(self, num_bytes: int) -> bytes:
        t0 = time()
        while len(self._buf) < num_bytes:
            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

            try:
                r = bytearray(self._port.read())
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        data = self._buf[:num_bytes]
        self._buf = self._buf[num_bytes:]
        return data

    def recv_until(self, bs: bytes) -> bytes:
        t0 = time()
        while True:
            try:
                i = self._buf.index(bs)
            except ValueError:
                pass
            else:
                break

            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

            try:
                r = bytearray(self._port.read())
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        i += 1
        data = self._buf[:i]
        self._buf = self._buf[i:]

        return data

    def send(self, data: bytes) -> None:
        self._port.write(data)

    def disconnect(self) -> None:
        self._port.close()
        self._port = None


class SerialProcessLink(BaseSerialLink):
    def __init__(self, port: Optional[int] = None) -> None:
        super().__init__()
        self._port = port
        self._process: Optional[mp.Process] = None

    def connect(self) -> None:
        self._recv_queue: "mp.Queue[bytes]" = mp.Queue()
        self._send_queue: "mp.Queue[Union[Tuple[str, int], bytes]]" = mp.Queue()
        self._flow_event = mp.Event()
        self._error_event = mp.Event()

        args = (
            self._port,
            self._baudrate,
            self._recv_queue,
            self._send_queue,
            self._flow_event,
            self._error_event,
        )

        self._process = mp.Process(
            target=serial_process_program,
            args=args,
            daemon=True,
        )

        if self._process is None:
            raise LinkError("failed to create subprocess")

        self._process.start()

        flow_event_was_set = self._flow_event.wait(self._timeout)

        if flow_event_was_set:
            log.debug("connect - flow event was set")
        else:
            log.debug("connect - flow event was not set (timeout)")
            self.disconnect()
            raise LinkError("failed to connect, timeout")

        if self._error_event.is_set():
            log.debug("connect - error event was set")
            self.disconnect()
            raise LinkError("failed to connect, see traceback from serial process")

        log.debug("connect - successful")

        self._buf = bytearray()

    def recv(self, num_bytes: int) -> bytes:
        self.__empty_queue_into_buf()

        t0 = time()
        while len(self._buf) < num_bytes:
            self.__get_into_buf()

            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

        data = self._buf[:num_bytes]
        self._buf = self._buf[num_bytes:]
        return data

    def recv_until(self, bs: bytes) -> bytes:
        self.__empty_queue_into_buf()

        n = len(bs)
        si = 0
        t0 = time()
        while True:
            buf_size = len(self._buf)
            if buf_size >= n:
                try:
                    i = self._buf.index(bs, si)
                except ValueError:
                    pass
                else:
                    break

                si = buf_size - n + 1

            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

            self.__get_into_buf()

        data = self._buf[: i + n]
        self._buf = self._buf[i + n :]
        return data

    def send(self, data: bytes) -> None:
        self._send_queue.put(data)

    def disconnect(self) -> None:
        if self._process is not None:
            if self._process.exitcode is None:
                self._flow_event.clear()
                self._process.join(1)

            if self._process.exitcode is None:
                log.info("forcing disconnect...")
                self._process.terminate()
                self._process.join(1)
                log.info("forced disconnect successful")

            if self._process.exitcode is None:
                raise LinkError("failed to disconnect")

    def __empty_queue_into_buf(self) -> None:
        while True:
            try:
                data = self._recv_queue.get_nowait()
            except queue.Empty:
                break

            self._buf.extend(data)

    def __get_into_buf(self) -> None:
        try:
            data = self._recv_queue.get(timeout=self._timeout)
        except queue.Empty:
            raise LinkError("recv timeout")
        except InterruptedError:
            return  # fixes interrupt issue on Windows

        self._buf.extend(data)

    @property
    def baudrate(self) -> int:
        return self._baudrate

    @baudrate.setter
    def baudrate(self, new_baudrate: int) -> None:
        self._baudrate = new_baudrate

        if self._process is not None and self._process.exitcode is None:
            log.debug("Changing baudrate to {}".format(new_baudrate))
            self._send_queue.put(("baudrate", new_baudrate))


def serial_process_program(
    port: int,
    baud: int,
    recv_q: "mp.Queue[bytes]",
    send_q: "mp.Queue[Union[Tuple[str, int], bytes]]",
    flow_event: th.Event,
    error_event: th.Event,
) -> None:
    log.debug("serial communication process started")
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    try:
        _serial_process_program(port, baud, recv_q, send_q, flow_event, error_event)
    except Exception:
        error_event.set()
        flow_event.set()
        sleep(0.1)  # give the main process some time to print log messages
        print("Exception raised in serial process:\n")
        traceback.print_exc()
        print("\n\n")

    recv_q.close()
    send_q.close()


def _serial_process_program(
    port: int,
    baud: int,
    recv_q: "mp.Queue[bytes]",
    send_q: "mp.Queue[Union[Tuple[str, int], bytes]]",
    flow_event: th.Event,
    error_event: th.Event,
) -> None:
    ser = serial.Serial(port=port, baudrate=baud, timeout=0, exclusive=True)
    flow_event.set()
    while flow_event.is_set():
        received = bytearray()
        while True:
            data = bytearray(ser.read(4096))
            if len(data) == 0:
                break
            received.extend(data)
        if len(received) > 0:
            recv_q.put(received)

        sent = False
        while True:
            try:
                x = send_q.get_nowait()
            except queue.Empty:
                break

            if isinstance(x, tuple):
                _, val = x  # assume its a baudrate change
                ser.baudrate = val
            else:
                ser.write(x)
                sent = True

        if not sent:
            sleep(0.0025)

    ser.close()



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import typing as t

import attrs

from PySide6.QtWidgets import QVBoxLayout

from acconeer.exptool import a121
from acconeer.exptool.a121.algo.distance._configs import get_high_accuracy_detector_config
from acconeer.exptool.a121.algo.distance._detector import Detector, DetectorConfig
from acconeer.exptool.a121.algo.distance._pidget_mapping import get_pidget_mapping
from acconeer.exptool.a121.model import power
from acconeer.exptool.app.new.ui.plugin_components import AttrsConfigEditor, GroupBox, pidgets
from acconeer.exptool.app.new.ui.resource_tab.event_system import EventBroker
from acconeer.exptool.app.new.ui.utils import LayoutWrapper, ScrollAreaDecorator, TopAlignDecorator

from ._preset_selection import create_plugin_selection_widget


@attrs.frozen
class DistanceConfigEvent:
    service_id: str
    config: DetectorConfig
    lower_power_state: t.Optional[power.Sensor.LowerPowerState]

    @property
    def translated_session_config(self) -> a121.SessionConfig:
        (session_config, _) = Detector._detector_to_session_config_and_processor_specs(
            self.config,
            sensor_ids=[1],
        )
        return session_config


class DistanceConfigInput(ScrollAreaDecorator):
    INTERESTS: t.ClassVar[set[type]] = set()
    description: t.ClassVar[str] = (
        "Specify distance configuration as in the Stream tab.\n\n"
        + "Additionally, you can specify a lower power state."
    )

    def __init__(self, broker: EventBroker, initial_config: DetectorConfig) -> None:
        layout = QVBoxLayout()
        super().__init__(TopAlignDecorator(LayoutWrapper(layout)))

        self.setMinimumWidth(400)

        self._broker = broker

        self.power_state_selection = pidgets.ComboboxPidgetFactory[
            t.Optional[power.Sensor.LowerPowerState]
        ](
            name_label_text="Lower power state:",
            items=[
                ("Don't use", None),
                ("Hibernate", power.Sensor.PowerState.HIBERNATE),
                ("Off", power.Sensor.PowerState.OFF),
            ],
        ).create(
            self
        )

        self.power_state_selection.set_data(None)
        self.power_state_selection.sig_update.connect(self._offer_event_if_config_is_valid)

        self.editor = AttrsConfigEditor(
            title="Distance config",
            factory_mapping=get_pidget_mapping(),
            config_type=DetectorConfig,
        )
        self.editor.set_data(initial_config)
        self.editor.sig_update.connect(self.editor.set_data)
        self.editor.sig_update.connect(self._display_validation_results)
        self.editor.sig_update.connect(self._offer_event_if_config_is_valid)

        wrapped_power_state_selection = GroupBox.vertical(left_header="")
        wrapped_power_state_selection.layout().addWidget(self.power_state_selection)

        preset_selection = create_plugin_selection_widget(
            "Presets",
            (
                "Balanced",
                [
                    lambda: self.editor.set_data(DetectorConfig()),
                    self._offer_event_if_config_is_valid,
                ],
            ),
            (
                "High accuracy",
                [
                    lambda: self.editor.set_data(get_high_accuracy_detector_config()),
                    self._offer_event_if_config_is_valid,
                ],
            ),
        )

        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(wrapped_power_state_selection)
        layout.addWidget(self.editor)
        layout.addWidget(preset_selection)

        (self.uninstall_function, self._id) = broker.install_identified_service(self, "distance")
        self._offer_event_if_config_is_valid()
        self.window_title = f"<b><code>[{self._id}]</code></b> Distance config"

    def _display_validation_results(self) -> None:
        config = self.editor.get_data()

        if config is None:
            return

        self.editor.handle_validation_results(config._collect_validation_results())

    def _offer_event_if_config_is_valid(self) -> None:
        if not self.editor.is_ready:
            return

        config = self.editor.get_data()

        if config is None:
            return

        try:
            config.validate()
        except a121.ValidationResult:
            return

        self._broker.offer_event(
            DistanceConfigEvent(
                self._id,
                config,
                self.power_state_selection.get_data(),
            )
        )

    def handle_event(self, event: t.Any) -> None:
        pass
